<?php

class KB_array extends KB_class {


	public function any_empty ($array = []) {

		if (is_empty($array))
			return true;

		foreach ($array as $value) {

			if (is_array($value) && is_empty($this->trim($value)))
				return true;

			else if (!is_array($value) && is_empty($value))
				return true;
		}

		return false;

	}


	public function complement ($array1 = [], $array2 = []) {

		if (any_empty(is_array($array1),is_array($array2)))
			return either($array1,$array2);

		$return = [];

		if (are_indexed($array1,$array2)) {

			foreach ($array1 as $value) {

				if (!array_match($value, $array2))
					$return[] = $value;
			}

		}

		else {

			foreach ($array1 as $key => $value) {
				if (!array_match(array($key => $value), $array2))
					$return[$key] = $value;
			}

		}

		return $return;

	}


	public function concat () {

		$arrays = func_get_args();

		if (count($arrays) < 2)
			return key_of(0,$arrays);

		$return = [];


		foreach ($arrays as $array) {

			if (is_empty($array))
				continue;

			if (is_empty($return))
				$return = $array;

			else
				$return = array_merge($return, (array) $array);
		}

		// foreach ($arrays as $array) {
			// $array = (array) $array;
		// }
		// $return = call_user_func_array("array_merge", $arrays);

			// $return = $return + (array) $array;

			// foreach((array) $array as $row)
				// $return[] = $row;

		return $return;

	}


	public function copy ($array=false) {
		//FAIL
		if (!$array || !is_array($array))
			return $array;

		//EXECUTE
		$copy = [];
		foreach($array as $key => $value) {
			if (is_array($value))
				$copy[$key] = $this->copy($value);
			else
			$copy[$key] = $value;
		}
		//RETURN
		return $copy;
	}


	public function create ($keys = [], $values = []) {

		if (any_empty($keys, $values, is_indexed($keys, true), is_indexed($values, true)))
			return [];

		$array = [];

		foreach ($keys as $index => $key)
			$array[$key] = key_of($index, $values);

		return $array;

	}


	public function difference ($array1 = [], $array2 = []) {

		if (any_empty($array1,$array2))
			return either($array1,$array2);

		/*
		$intersection = $this->intersect($array1, $array2);
		$union = $this->merge($array1, $array2);
		$return = $this->complement($union,$intersection);
		*/

		$part1 = $this->complement($array1,$array2);
		$part2 = $this->complement($array2,$array1);
		$return = $this->merge($part1,$part2);

		return $return;

	}


	public function exclude ($array = [], $filter = false) {

		$include = $this->filter($array, $filter);
		return $this->difference($array, $include);

	}


	public function exclude_by ($array = [], $key = false, $filter = false) {

		$include = $this->filter_by($array, $key, $filter);
		return array_diff_assoc($array, $include);

	}


	public function filter ($array = [], $filter = false) {

		if (is_empty($filter))
			return $array;

		if (are_empty(is_string($filter), is_array($filter)))
			return $array;

		if (any_empty($array, is_array($array)))
			return [];

		$map = true;

		foreach (array_values($array) as $value) {
			if (!is_array($value)) {
				$map = false;
				break;
			}
		}

		if ($map) {

			$return = [];

			foreach ($array as $key => $value)

				if (not_empty($callback = $this->filter($value, $filter)))
					$return[$key] = $callback;

			return $return;

		}

		if (is_string($filter) && in_list($filter, array_values($array)))
			return $array;

		if (is_array($filter)) {

			if (is_empty(keys_of(array_keys($filter), $array)))
				return [];

			foreach ($filter as $key => $value) {

				if (!is_string($key))
					return [];

				if (is_array($value) && in_list(key_of($key, $array), array_values($value)))
					continue;

				if (!in_list($value, array_values(keys_of($key, $array)))) {

					if (has_regex($value) && not_empty(fetch([$key => $value], $array)))
						continue;

					return [];

				}
			}

			return $array;
		}

		return [];
	}


	public function filter_by ($array = [], $key = false, $filter = false) {

		// FAIL
		if (is_empty($array) || (!$key && !$filter))
			return $array;

		// EXECUTE
		if (is_string($key) && not_empty($filter)) {
			$array = fetch(array($key=>$filter),$array);
			return $array ? $array : [];
		}

		else if (is_array($key) && !$filter) {
			foreach($key as $filter_key => $filter_value)
				if (not_empty($filter_value))
					$array = array_find(array($filter_key=>$filter_value),$array,"^","$","im");
			return $array;
		}

		else if (is_array($key) && $filter) {
			foreach($key as $filter_key => $filter_value)
				$array = array_find(array($filter_key=>array($filter_value=>$filter)),$array,"^","$","im");
			return $array;
		}


		return $array;
	}


	public function find ($array, $find) {
		return array_find($find, $array);
	}


	public function fill ($array = [], $new_key = "", $set_value = "") {

		if (any_empty($array, is_array($array), $new_key))
			return $array;

		if (is_array($new_key)) {

			foreach ($new_keys = $new_key as $new_key => $set_value)
				$array = $this->fill($array, $new_key, $set_value);

			return $array;
		}

		foreach ($array as $key => $value) {

			if (is_array($value)) {
				$array[$key] = $this->fill($value, $new_key, $set_value);
				continue;
			}

			if (!array_key_exists($new_key, $array))
				$array[$new_key] = $set_value;

			break;
		}

		return $array;

	}


	public function flip_index ($array = []) {

		if (any_empty($array, is_array($array)))
			return $array;

		$flipped_array = [];

		foreach ($array as $key => $entry) {

			if (!is_array($entry))
				return $array;

			foreach ($entry as $index => $value) {

				$flipped_array[$index][$key] = $value;
			}

		}

		return $flipped_array;
	}


	public function get_highest_group_key ($array = false, $group = false) {

		deprecated();

		//FAIL
		if ((!is_array($array) || empty($array)) && !is_object($array))
			return false;

		if (!$group)
			return false;

		// PREPARE

		$compare = $this->group_by($array,$group);
		$group_key = false;

		foreach(array_keys($compare) as $key) {

			if (!isset($array[$key])) {
				$group_key = $key;
				break;
			}

		}
		return $group_key;
	}


	public function group_by ($array = false, $group = false, $groups = false) {

		//FAIL
    if (!$group)
      return $array;

		if ((!is_array($array) || empty($array)) && !is_object($array))
			return $array;

		if (!array_key_exists($group,$array[key($array)]) && !isset($array[key($array)]->$group))
			return $array;

		//PREPARE
		$return = [];
		$return2 = [];

		//EXECUTE

		foreach($array as $key => $entry)
			if (array_key_exists($group,$entry))
			 	$return[$entry[$group]][$key] = $entry;

		if (not_empty($groups))
			deprecated("Using groups as filter is deprecated");

		if (is_array($groups)) {
			$groups = $groups+array("");

			$remove_keys = $this->diff(
				$groups,
				array_keys($return)
			);
			if (not_empty($remove_keys))
				$return = remove_keys_of($remove_keys, $return);

		}

		//RETURN
		return $return;

	}


	public function has_regex ($array = []) {

		if (any_empty($array, is_array($array)))
			return false;

		foreach ($array as $value) {

			if (is_array($value) && $this->has_regex($value))
				return true;

			else if (has_regex($value))
				return true;

		}
		return false;
	}


	public function hierarchie ($array = false, $key = false, $parent = false, $marker = false, $entry_point = false, $depth = 0) {

		if (any_empty($array, $key, $parent))
			return [];

		if (any_empty(key_of($key, current($array)), array_key_exists($parent, current($array))))
			return [];

		if ($entry_point === false) {

			$parents = [];
			$keys = [];

			foreach ($array as $index => $value) {

				$keys[] = key_of($key, $value);
				$parents[] = key_of($parent, $value);

			}

			foreach ($parents as $value)
				if (!$this->match($value, $keys))
					return $this->hierarchie($array, $key, $parent, $marker, $value, $depth);


			return [];

		}

		$return = [];

		foreach ($array as $index => $value) {

			if (are_equal(key_of($parent, $value), $entry_point)) {

				if (not_empty($marker) && is_string($marker))
					$value[$marker] = $depth;

				unset($array[$index]);

				$return[$index] = $value;
				$return = $this->union($return, $this->hierarchie($array, $key, $parent, $marker, key_of($key, $value), $depth+1));
			}
		}

		return $return;

	}


	public function highest_group ($array = [], $parent = "parent_ID", $key = "ID") {

		if (is_empty($array))
			return false;

		$parents = [];
		$IDs = [];

		foreach ($array as $entry) {

			if (key_of($parent, $entry) === NULL)
				return "";

			if (key_of($key, $entry) === NULL)
				return false;

			$parents[] = key_of($parent, $entry);
			$IDs[] = key_of($key, $entry);

		}

		$highest_group = $this->complement($parents, $IDs);

		if (!is_array($highest_group))
			return false;

		return current($highest_group);

	}


	public function implode ($array = false, $columns = false, $options = []) {

		if (!is_array($array))
			return $array;

		// Special case: implode everyting

		if ($columns === false) {
			foreach ($array as $key => &$value) {
				if (is_array($value))
					$value = $this->implode($value);
			}
			$array = implode($array);
		}
		else{

			// FAIL
			if (is_empty($array))
				return $array;

			foreach($array as $key => $value) {


				if (count((array) $columns) > 1) {

					unset($array[$key]);

					foreach ((array) $columns as $column)
						$array[$key][] = key_of($column, $value) . key_of("separator",key_of($column,$options));

					$array[$key] = implode($array[$key]);

				}

				else {

					foreach ((array) $columns as $column)
						$array[$key] = key_of($column, $value);
				}

			}
		}

		return $array;
	}


	public function index ($array = [], $index = "") {

		if (any_empty($array, is_array($array), $index, is_string($index)))
			return $array;

		if (!is_array(current($array)))
			return $array;

		$return = [];

		foreach ($array as $key => $value) {

			if (not_empty($index_key = key_of($index, $value)))
				$return[$index_key] = $value;

			unset($array[$key]);

		}

		return $return;

	}


	public function intersect ($array1 = false, $array2 = false) {

		if (are_empty($array1,$array2))
			return [];

		if (is_array($array1) && !is_array($array2))
			return $array1;

		if (is_array($array2) && !is_array($array1))
			return $array2;

		$return = [];

		if (are_indexed((array) $array1, (array) $array2)) {

			foreach ((array) $array1 as $value) {

				if (!is_array($value) && in_array($value, $array2))
				 	$return[] = $value;

				else if (array_match($value, $array2, false, false))
					$return[] = $value;

			}

		}

		else {

			foreach ((array) $array1 as $key => $value) {
				if (array_match(array($key => $value), $array2, false, false))
					$return[$key] = $value;
			}

		}

		return $return;

	}


	public function match ($needle = [], $haystack = []) {

		if (are_empty($needle, $haystack))
			return true;

		if (any_empty($needle, $haystack, is_array($haystack)))
			return false;

		if (is_array($needle) && is_string($haystack))
			return false;

		if (is_scalar($needle))
			$needle = array($needle);

		if (are_equal(serialize($needle), serialize($haystack)))
			return true;

		if (is_array(current($needle))) {

			foreach ($needle as $key => $value) {

				$array = key_fetch($key, $haystack);

				if ((is_string($value) || is_int($value)) && !are_equal($value, $array))
					return false;

				if (is_array($value) && !$this->match($value, $array))
					return false;

			}
			return true;
		}

		else if (is_array(current($haystack))) {

			foreach ($haystack as $array) {
				if ($this->match($needle, $array))
					return true;
			}

			return false;

		}

		else if (count($haystack) > 1) {

			if (is_indexed($haystack)) {
				foreach ($haystack as $key => $value) {
					if ($this->match($needle, array($value)))
						return true;
				}
			}

			else {
				foreach ($haystack as $key => $value) {
					if ($this->match($needle, array($key=>$value)))
						return true;
				}

			}

			return false;

		}

		else if (are_equal(key($needle), key($haystack)) && are_equal(current($needle), current($haystack)))
			return true;

		return false;
	}


	public function move ($array = [], $from = false, $to = false) {

		if (any_empty($array, $from, $to, is_array($array), is_numeric($from), is_numeric($to)))
			return (array) $array;

		array_splice($array, $to, 0, array_splice($array, $from, 1));

		return $array;

	}


	public function pop ($array = NULL) {

		if (any_empty($array, is_array($array)))
			return "";

		return current(array_slice($array,-1));
	}


	public function range_by ($array = [], $key = false, $range = false) {

		if (any_empty($array, $key, $range))
			return $array;

		if (!is_array($range))
			$range = array($range,"");

		if ($key === true) {

			$start = false;
			$end = false;
			$array_cc = $this->copy($array);

			ksort($array_cc);

			foreach ($array_cc as $key => $value) {

				if (!$start && $key >= $range[0])
					$start = true;

				if (!$start || $end)
					unset($array_cc[$key]);

				if ($start && !$end && $key == $range[1])
					$end = true;

			}

			$array = array_intersect_key($array,$array_cc);

			return $array;
		}


		$array_sorted = $this->copy($array);


		if ($range[0] != "")
			$array_sorted = array("range_start" => array($key => $range[0])) + $array_sorted;

		if (isset($range[1]) && $range[1] != "")
			$array_sorted = $array_sorted + array("range_end" => array($key => $range[1]));

		$array_sorted = $this->sort_by($array_sorted,$key);

		if (isset($array_sorted["range_start"]))
			$bool = false;

		else
			$bool = true;



		foreach ($array_sorted as $index => $value) {

			$index = strval($index);

			if ($index == "range_start") {
				unset($array_sorted[$index]);
				$bool = true;
				continue;
			}

			if ($index == "range_end")
				$bool = false;

			if (!$bool)
				unset($array_sorted[$index]);
		}

		$array = remove_keys_of(
			$this->difference(
				array_keys($array_sorted),
				array_keys($array)
			),
			$array
		);

		return $array;

	}


	public function render ($array = [], $depth = 0) {

		if (is_object($array))
			return get_class($array); //var_export($array, true);

		if ($array === true)
			return "true";

		if ($array === false)
			return "false";

		if ($array === NULL)
			return "NULL";

		if (is_string($array))
			return '"'.$array.'"';

		if (is_int($array))
			return $array;

		if (is_float($array))
			return $array;

		if (is_array($array)) {

			$return = "array(\n";

			$padding_outer = str_repeat("\t",$depth);
			$padding_inner = str_repeat("\t",$depth+1);

			foreach ($array as $key => $value) {

				$return .= $padding_inner;

				if (is_int($key))
					$return .= $key;
				else
					$return .= '"'.$key.'"';

				$return .= " => ";

				$return .= $this->render($value,$depth+1);

				$return .= ",\n";
			}
			$return .= $padding_outer;
			$return .= ")";

			return $return;
		}
	}


	public function shift ($array = NULL) {

		if (any_empty($array, is_array($array)))
			return "";

		reset($array);
		return current($array);
	}


	public function shuffle ($array = []) {

		if (any_empty($array, is_array($array)))
			return (array) $array;

		$keys = array_keys($array);
		shuffle($keys);

		$return = [];

		foreach ($keys as $key)
			$return[$key] = $array[$key];

		return $return;
	}


	public function sort ($array, $sorting="", $case_insensitive=true) {
		//FAIL
		if (!is_array($array) && !is_object($array))
			return $array;

		//CONFIG
		if ($case_insensitive == true)
			$strcmp_fn = "strnatcasecmp";
		else
			$strcmp_fn = "strnatcmp";

		//EXECUTE
		asort($array);

		if ($sorting == "DESC")
			$array = array_reverse($array);

		//RETURN
		return $array;
	}


	public function sort_by ($array, $sort_by = 'name', $sorting="", $case_insensitive=true) {

		if (are_empty(is_array($array), is_object($array)))
			return $array;

		if (is_array($array) && !is_array(current($array)))
			return $array;

		if (!is_array($sort_by))
			$sort_by = [$sort_by => false];

		if (are_empty(array_key_exists(key($sort_by), current($array)), isset($array[key($array)]->{key($sort_by)})))
			return $array;

		//CONFIG

		if ($case_insensitive === true)
			$strcmp_fn = function($a, $b) { return strnatcasecmp(either($a,""), either($b,""));};

		else
			$strcmp_fn = function($a, $b) { return strnatcmp(either($a,""), either($b,""));};



		//PREPARE

		// Quicksort Bugfix with Schwartzian transform
		// Add Decoration

		$i = 0;

		if (is_array($array))
			foreach($array as &$entry)
				$entry[key($sort_by)] = array($entry[key($sort_by)],$i++);

		if (is_array($array))
			$fn = function($a, $b) use($strcmp_fn, $sort_by) {

				if (is_array($a) && is_array($b)) {

					if (not_empty(current($sort_by))) {

						if (
							(
								$return = $strcmp_fn(
									key_of(current($sort_by), $a[key($sort_by)][0]),
									key_of(current($sort_by), $b[key($sort_by)][0])
								)
							) == 0
						)
							return $strcmp_fn($a[key($sort_by)][1], $b[key($sort_by)][1]);

						else
							return $return;
					}
					else{

						if (($return = $strcmp_fn($a[key($sort_by)][0], $b[key($sort_by)][0])) == 0)
						 	return $strcmp_fn($a[key($sort_by)][1], $b[key($sort_by)][1]);

						else
							return $return;
					}
				}
				else
					return 0;
			};

		else if (is_object($array))
			$fn = function($a, $b) use($strcmp_fn, $sort_by) {

				if (is_object($a) && is_object($b)) {

					if (not_empty(current($sort_by))) {
						return $strcmp_fn($a->{key($sort_by)}->{current($sort_by)}, $b->{key($sort_by)}->{current($sort_by)});
					}
					else{
						return $strcmp_fn($a->{key($sort_by)}, $b->{key($sort_by)});
					}
				}
				else
					return 0;
			};

		//EXECUTE
		uasort($array, $fn);

		// Quicksort Bugfix with Schwartzian transform
		// Remove Decoration

		if (is_array($array))
		foreach($array as &$entry) {
			$entry[key($sort_by)] = $entry[key($sort_by)][0];
		}

		if ($sorting == "DESC")
			$array = array_reverse($array,true);

		//RETURN
		return $array;

	}


	public function sort_by_hierarchie ($array = [], $groups = [], $key = false, $group_key = false, $marker = false) {

		if (any_empty($array, is_array($array)))
			return [];

		if (any_empty($groups, $key, $group_key))
			return $array;

		// if (any_empty(key_of($key, current($array)), key_of($key, current($groups)), array_key_exists($group_key, current($array))))
		if (any_empty(array_key_exists($group_key, current($array)), array_key_exists($key, current($groups))))
			return $array;

		$return = [];

		foreach ($groups as $group) {

			foreach ($array as $index => $value) {

				if (are_equal(key_of($group_key, $value), key_of($key, $group))) {

					unset($array[$index]);

					if (not_empty($marker) && is_string($marker))
						$value[$marker] = key_of($marker, $group);

					$return[$index] = $value;

				}
			}
		}

		return $return;

	}


	public function trim ($array = []) {

		if (!is_array($array))
			return [];

		foreach ($array as $key => $value) {

			if (is_array($value) && is_empty($this->trim($value)))
				unset($array[$key]);

			else if (!is_array($value) && is_empty($value))
				unset($array[$key]);

		}

		return $array;

	}


	public function ungroup ($array = false, $group_value = NULL, $key = "ID", $depth = 0) {
		// Ungroup for non indexed arrays

	 	if (is_empty($array))
	 		return $array;

	 //PREPARE
	 $return = [];

	 // Execute and return {


	 if (is_array($group_value)) {

		 if (is_indexed($group_value, true)) {

			 foreach ($group_value as $value) {
			 		$groups[key_of($key, $value)] = $value;
			 }

			 $group_value = $groups;

		 }

		 $group_value = array(
			 ""=>array(
				 "depth" => "0"
			 )
		 )+$group_value;

		 foreach($group_value as $group_key => $group_value) {
			 if (array_key_exists($group_key,$array))
			 foreach($array[$group_key] as $key => $value) {
			 // while(isset($array[$group_value][$i])) {

				 $return[$key] = $value;
				 $return[$key]["depth"] = key_fetch("depth",$group_value);
				 // $next = $key;

				 // if (isset($array[$next]))
					 // $return = $return + $this->ungroup($array, $next, $depth);

			 }
		 }

	 }
	 else{

		 foreach($array[$group_value] as $key => $value) {
		 // while(isset($array[$group_value][$i])) {

			 $return[$key] = $value;
			 $return[$key]["depth"] = $depth;
			 $next = $key;

			 if (isset($array[$next]))
				 $return = $return + $this->ungroup($array, $next, $key, $depth+1);

		 }
	 }
	 return $return;
	 //}
	 }


	public function ungroup_by ($array = false, $group = false, $group_value = "", $depth = -1) {

		if (is_empty($group)) {
			return $this->ungroup($array, $group_value, "ID", $depth);
		}

		if ((!is_array($array) || empty($array)) && !is_object($array))
			return $array;

		$return = [];
		$depth++;

		if (is_empty(key_of($group_value, $array)))
			return [];

		foreach ($array[$group_value] as $value) {
			$value["depth"] = $depth;
			$return[] = $value;
			$return = $this->merge($return, $this->ungroup_by($array, $group, key_of($group, $value), $depth));
		}

		return $return;

	}


	public function union () {

		foreach ($arrays = func_get_args() as $index => $array) {
			if (!is_array($array))
				unset($arrays[$index]);
		}

		if (is_empty($arrays))
			return [];

		$arrays = array_values($arrays);

		if (count($arrays) < 2)
			return key_of(0,$arrays);

		// $arrays = array_reverse($arrays);

		$count = count($arrays);

		for ($i=1; $count, $i < $count; $i++) {

			if (are_indexed($arrays[0], $arrays[$i]))
				$arrays[0] = array_merge(
					$arrays[0],
					$this->complement($arrays[$i],$arrays[0])
				);
				// $arrays[0] = array_merge($arrays[0],$arrays[$i]);

			else
				$arrays[0] = array_multimerge($arrays[0],$arrays[$i]);

		}

		return $arrays[0];
	}


	public function unique ($array = [], $key = NULL) {

		if (is_object($array))
			return $array;

		if (!is_array($array))
			return array($array);

		if (is_null($key))
			return array_unique($array);

		if (!is_string($key) && !is_indexed($array))
			return $array;

		$return = [];
		$list = [];

		foreach ($array as $index => $entry) {

			if (in_list($key_value = key_of($key, $entry), $list))
				continue;

			$list[] = $key_value;
			$return[$index] = $entry;

		}

		return $return;

	}




	// Alias


	public function first ($array = NULL) {

		return $this->shift($array);

	}

	public function merge () {


		return call_user_func_array(array($this,"union"),func_get_args());

	}

	public function last ($array = NULL) {

		return $this->pop($array);

	}


	// Backwards compatibility

	public function diff ($array1 = [], $array2 = []) {

		return $this->complement($array1, $array2);

	}
}
