<?php

class KB_db extends PDO {

	public $table;

	private
		$calc_time,
		$config,
		$core,
		$dev,
		$db_data,
		$db,
		$error_info,
		$queries,
		$time_end,
		$time_start,
		$times;

	protected $transaction_active = false;
	private $no_string = ["CURRENT_DATE()","NOW()","NULL",NULL];
	private $no_equal = array("NULL");

  function __construct($core, $db_data){

		$this->core = $core;

		// Deprecated
		$this->db_data = $db_data;

		$this->config($db_data);
		$this->table = new KB_db_table($core);
		$this->table->config($this->config());

  	global $is_test_server;
  	global $dev_prefix;

		$this->times = array(0,0);
		$this->calc_time = false;
		$this->core = $core;
		$this->error_info = array("","","");


		if (isset($db_data['short']) && !empty($db_data['short']) && $is_test_server){
			$db = $dev_prefix.$db_data['short'];
			$host = "localhost";
		}
		else{
			$this->db = $db_data['database'];
			$host = $db_data['server'];
		}

    parent::__construct('mysql:host='.$host.';dbname='.$this->db, $db_data['user'], $db_data['password']);
    $this->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
		$this->setAttribute(PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, true);

		if (intval(preg_replace("/\./","",phpversion())) > 512)
			$this->setAttribute(PDO::ATTR_EMULATE_PREPARES, true);

		$this->query("SET NAMES 'utf8';");
		$this->query("SET lc_time_names = 'de_DE';");
		//$this->query("SET CLIENT_ENCODING TO 'UTF8';");
		//$this->query("SET CHARACTER SET utf8");

  }


	// PUBLIC
	public function config($name = NULL, $value = NULL) {

		if (is_empty($name))
			return $this->config;

		if (is_array($name) && is_null($value)) {
			foreach ($name as $var => $value) {

				if (!is_scalar($var))
					continue;

				$this->config($var, $value);
			}
			return;
		}

		if (!is_string($name))
			return;

		if (is_null($value))
			return key_of($name, $this->config);

		$this->config[$name] = $value;

	}

	public function delete($table = "", $search = [], $db = "") {

		if (any_empty($table, $search))
			return false;

		if (!$this->table->exists($table, $db))
			return true;

		$sql = "DELETE FROM ".ifttt($db,$db.".").$this->config("prefix").$table.$this->where($search).";";
		return $this->sql_query($sql, "Delete from $table");
	}


	public function dump($options = []) {

		$options = $this->core->array->merge(
			array(
				"add_drop_table" => false,
				"single_transaction" => true,
				"add_locks" => true,
				"complete_insert" => true,
				"extended_insert" => true,
			),
			$options
		);

		$commands = [];

		if (key_of("add_drop_table",$options))
			$commands[] = "--add-drop-table";

		if (key_of("single_transaction",$options))
			$commands[] = "--single-transaction";

		if (key_of("add_locks",$options))
			$commands[] = "--add-locks";

		if (key_of("complete_insert",$options))
			$commands[] = "--complete-insert";

		if (key_of("extended_insert",$options))
			$commands[] = "--extended-insert";

		$commands[] = key_of("database",$this->db_data);
		$commands[] = "--no-create-db";
		$commands[] = "--user=".key_of("user",$this->db_data);
		$commands[] = "--password=".key_of("password",$this->db_data);
		$commands[] = "--host=".key_of("server",$this->db_data);
		// $commands[] = "--databases ".key_of("database",$this->db_data);

		$return = exec("mysqldump ".implode(" ",$commands), $out, $ret);
		return implode("\n",$out);

	}


	public function end_transaction() {

  	if ($this->transaction_active && $this->core->message->set($this->commit(), "DB - Commiting Transaction"))
  		$this->transaction_active = false;

  }


	public function found_rows(){

		$this->time();

		try {
			$rs1 = $this->query('SELECT FOUND_ROWS()');
			$rowCount = (int) $rs1->fetchColumn();
		}

		catch (PDOException $e) {
  		$this->error($e,$sql,$sth);
		}

		$this->time('SELECT FOUND_ROWS()');
  	return $rowCount;
  }


	public function insert($table = "", $columns = [], $values = [], $db = "") {

		if (any_empty($table, $columns))
			return false;

		if (is_indexed($columns) && is_empty($values))
			return true;

		// Associated array

		if (is_empty($values) && !is_indexed($columns)) {

			$values = [];

			// $values = array_values($columns);
			foreach (array_values($columns) as $value) {

				if (is_array($value))
					$value = $this->core->json->encode($value);

				if (!in_list($value,$this->no_string))
					$value = "'".addslashes($value)."'";

				if (is_null($value))
					$value = "NULL";

				$values[] = $value;
			}

			$columns = array_keys($columns);
			$columns = "(`".implode("`,`", $columns)."`)";
			$values = "(".implode(",", $values).")";

		}

		else if (is_indexed($columns) && is_indexed($values)) {

			$columns = "(".implode(",", $columns).")";

			foreach ($values as $row => $entry) {

				foreach ($entry as $column => $value) {

					if (is_array($value))
						$value = $this->core->json->encode($value);

					if (is_null($value))
						$value = "NULL";

					else
						$value = addslashes($value);

					if (!in_list($value,$this->no_string))
						$entry[$column] = "'".$value."'";

				}
				$values[$row] = "(".implode(",", $entry).")";
			}

			$values = implode(",", $values);

		}

		else {
			return false;
		}

		$sql = "INSERT INTO ".ifttt($db,$db.".").$this->config("prefix").$table." $columns VALUES $values;";
		return $this->sql_insert($sql, "Insert $table");

	}


	public function name() {
		return $this->db;
	}


	public function num_rows($sql, $message=false){
  	$this->time();
  	try{
  		$sth = $this->prepare($sql);
  		$var = $sth->execute();
  		($message) ? $this->core->message->set($var, $message." - ".$sql) : true;
  		$rows = count($sth->fetchAll());
  	} catch (PDOException $e) {
  		$this->error($e,$sql,$sth);
	  }
  	$this->time($sql);
  	return $rows;
  }


	public function reorder($table = "", $column = "", $group = false, $entry = []) {

		if (!$this->table->exists($table))
			return false;

		if (any_empty($table, $column))
			return false;

		// We have an entry

		if (not_empty($entry)) {

			if (is_empty($ID = key_of("ID", $entry)))
				return true;

			// Entry has a group
			if ($group) {

				$position = either(key_of($column, $entry),"NULL");

				// No group change
				if (is_string($group) && !array_key_exists($group,$entry)) {

					$sql = "
					SELECT @po := $column, @go := IFNULL($group,0) FROM ".$this->config("prefix").$table." WHERE ID = '".$ID."';
					SET @pn := $position;
					SET @c := 0;
					INSERT INTO ".$this->config("prefix").$table." (ID, $column)
					SELECT ID, c FROM (
						SELECT *, @c:=@c+1 AS c FROM (SELECT ID, $column, CASE WHEN $group IS NULL THEN 0 ELSE $group END AS $group FROM ".$this->config("prefix").$table.") AS ".$this->config("prefix").$table." WHERE $group IN (@go)
						ORDER BY
						CASE WHEN @po <= IFNULL(@pn,@po) THEN ($column < @po) END DESC,
						CASE WHEN @po > IFNULL(@pn,@po) THEN ($column < IFNULL(@pn,@po)) END DESC,
						CASE WHEN @po < IFNULL(@pn,@po) THEN ($column > @po && $column < @pn+1) END DESC,
						CASE WHEN @pn IS NOT NULL THEN ($column = @po) END DESC,
						CASE WHEN @po < IFNULL(@pn,@po) THEN ($column > @pn) END DESC,
						$column
						)
						AS sub ON DUPLICATE KEY UPDATE $column=VALUES($column);
						";

				}

				// New group is set
				else {

					if (!is_array($group))
						$group = array($group => key_of($group,$entry));

					$group_key = key($group);
					$group = current($group);

					$sql = "
					SELECT @po := $column, @go := IFNULL($group_key,0) FROM ".$this->config("prefix").$table." WHERE ID = '".$ID."';
					SET @pn := $position;
					SET @gn := ".ifttt($group === "NULL", "0", $group).";
					SET @c := 0;
					SET @g := 0;
					INSERT INTO ".$this->config("prefix").$table." (ID, $column)
					SELECT ID, c FROM (
						SELECT ID, $column, @c:=if (IFNULL($group_key,0) = @g, @c+1, if (@go!=@gn AND ID='".$ID."',@c+1,1)) AS c, @g := if (@go!=@gn AND ID='".$ID."',@g,IFNULL($group_key,0)) FROM ".$this->config("prefix").$table." WHERE $group_key IN (@go,@gn)".ifttt($group === "NULL"," OR parent_ID IS NULL")."
						ORDER BY
						CASE WHEN @go!=@gn THEN ($group_key = @gn) END DESC,
						CASE WHEN @go!=@gn THEN (ID = '".$ID."') END DESC,
						$group_key,
						CASE WHEN @go=@gn && @po <= IFNULL(@pn,@po) THEN ($column < @po) END DESC,
						CASE WHEN @go=@gn && @po > IFNULL(@pn,@po) THEN ($column < IFNULL(@pn,@po)) END DESC,
						CASE WHEN @go=@gn && @po < IFNULL(@pn,@po) THEN ($column > @po && $column < @pn+1) END DESC,
						CASE WHEN @go=@gn && @pn IS NOT NULL THEN ($column = @po) END DESC,
						CASE WHEN @go=@gn && @po < IFNULL(@pn,@po) THEN ($column > @pn) END DESC,
						$column
						)
						AS sub ON DUPLICATE KEY UPDATE $column=VALUES($column);
						";
				}

			}

			// Entry has no group
			else {

				if (is_empty($position = key_of($column, $entry)))
					return true;

				$sql = "
					SET @o := (SELECT $column FROM ".$this->config("prefix").$table." WHERE ID = '".$ID."');
					SET @n := $position;
					SET @c := 0;
					INSERT INTO ".$this->config("prefix").$table." (ID, $column) SELECT ID, c FROM (
						SELECT ID, $column, @c:=@c+1 AS c FROM ".$this->config("prefix").$table." ORDER BY
							CASE @o<@n WHEN 1 THEN ($column < @o) ELSE ($column < @n) END DESC, CASE @o<@n WHEN 1 THEN ($column > @o && $column < @n+1) ELSE NULL END DESC , ($column = @o) DESC , CASE @o<@n WHEN 1 THEN ($column > @n) ELSE NULL END DESC , $column
					) AS sub ON DUPLICATE KEY UPDATE $column=VALUES($column);
				";
			}
		}

		// We have no entry
		else {

			// We have a group

			if ($group) {

				// Group is specified
				if (is_array($group)) {

					$group_key = key($group);
					$group = current($group);

					$group_select = ifttt(
						$group === "NULL",
						"$group_key IS NULL",
						"$group_key = '$group'"
					);

					$sql = "
						SET @c := 0;
						INSERT INTO ".$this->config("prefix").$table." (ID, $column) SELECT ID, c FROM (
							SELECT ID, @c:=@c+1 AS c FROM ".$this->config("prefix").$table." WHERE $group_select ORDER BY ISNULL($column), $column
						) AS sub ON DUPLICATE KEY UPDATE $column=VALUES($column);
					";

				}

				// Group is not specified
				// Reorder all groups
				else {

					$sql = "
						SET @c := 0;
						SET @g := 0;
						INSERT INTO ".$this->config("prefix").$table." (ID, $column) SELECT ID, c FROM (
							SELECT ID, @c:=if (IFNULL($group,0) = @g, @c+1, 1) AS c, @g := IFNULL($group,0)  FROM ".$this->config("prefix").$table." ORDER BY $group, ISNULL($column), $column
						) AS sub ON DUPLICATE KEY UPDATE $column=VALUES($column);
					";

				}
			}

			// We have no group

			else {

				$sql = "
					SET @c := 0;
					INSERT INTO ".$this->config("prefix").$table." (ID, $column) SELECT ID, c FROM (
						SELECT ID, @c:=@c+1 AS c FROM ".$this->config("prefix").$table." ORDER BY ISNULL($column), $column
					) AS sub ON DUPLICATE KEY UPDATE $column=VALUES($column);
				";

			}
		}
		return $this->sql_query($sql,"Reorder $table");
	}


	public function rows($table = "", $where = []) {

		if (is_empty($table))
			return 0;

		$table = $this->config("prefix").$table;
		$where = $this->where($where);

		$sql = "SELECT count(*) FROM $table $where";

		return current($this->core->db->sql_first($sql, "Count $table"));

	}


	public function select($table = "", $where = [], $order = [], $limit = false) {

		$where = $this->where($where, array("table"=>$table));
		$table = $this->from($table);

		if (is_empty($table))
			return [];

		$order = $this->order_by($order);
		$limit = $this->limit($limit);

		$sql = "SELECT $table $where $order $limit";

		return $this->sql_all($sql);

	}


	public function select_first($table = "", $where = [], $order = []) {
		$where = $this->where($where, array("table"=>$table));

		$table = $this->from($table);

		if (is_empty($table))
			return [];

		$order = $this->order_by($order);

		$sql = "SELECT $table $where $order";

		return $this->sql_first($sql);

	}


	public function set_rollback() {

  	if ($this->transaction_active) {
  		$this->core->message->set(!$this->rollBack(), "DB - ROLLBACK");
  		$this->transaction_active = false;
  	}
  }


	public function showexectime() {

		if ($this->calc_time == true && $this->dev == true) {

			for($i=0;$i<count($this->time_end);$i++) {
				$this->times[1] += $this->time_end[$i][1] - $this->time_start[$i][1];
				$this->times[0] += $this->time_end[$i][0] - $this->time_start[$i][0];
			}

			return ($this->times[0]+$this->times[1]);
		}
		return false;
	}


	public function showquerytime() {

		if ($this->calc_time == true && $this->dev == true) {

			$dev = "<div style=\"background:gray;color:white;padding:10px;border:dashed 5px blue;margin:5px;\">";
			$dev .= "<table style=\"font-size:10px;\">";

			for($i=0;$i<count($this->time_end);$i++) {

				$this->times[1] = $this->time_end[$i][1] - $this->time_start[$i][1];
				$this->times[0] = $this->time_end[$i][0] - $this->time_start[$i][0];
				$dev .= "<tr><td nowrap>No #".$i.": </td><td nowrap>".($this->times[0]+$this->times[1])."</td><td>".$this->queries[$i]."</td></tr>";
			}

			$dev .= "<tr><td nowrap>SUM: </td><td nowrap>".$this->showexectime()."</td><td></td></tr>";
			$dev .= "</table></div>";

			return $dev;
		}

		return false;
	}


	public function sql_all($sql) {

  	$this->time();
  	$row = [];
    $this->core->db_call_count++;
    $this->core->db_call[] = $sql;

  	try {

			$sth = $this->prepare($sql);
	    $sth->execute();
    	$row = $sth->fetchAll(PDO::FETCH_ASSOC);
    }

		catch (PDOException $e) {
  		$this->error($e,$sql,$sth);
		  $this->error_info = $e->errorInfo;
		  return false;
	  }

  	$this->time($sql);

  	return $row;
  }


	public function sql_first($sql) {

		$this->time();
		$this->core->db_call_count++;
		$this->core->db_call[] = $sql;

		try {

			$sth = $this->prepare($sql);
			$sth->execute();
			$row = $sth->fetch(PDO::FETCH_ASSOC);
		}

		catch (PDOException $e) {
				$this->error($e, $sql,$sth);
		}

		if (empty($row))
			$row = [];

		$this->time($sql);
			return $row;
	}


	public function sql_insert($sql, $message=false) {

		$this->time();
    $this->core->db_call_count++;

		try {
    	$sth = $this->prepare($sql);
    	$var = $sth->execute();
    	$message = ($message) ? $this->core->message->set($var, $message." - ".$sql) : true;
    	$ID = $this->lastInsertId();

			if ($ID == 0)
				$ID = true;
  	}
		catch (PDOException $e) {
			$this->error_info = $e->errorInfo;
			$this->set_rollback();
			$ID = false;
			$message = $this->error($e,$sql,$sth);
		}
  	$this->time($sql);
  	return $ID;
  }


	public function sql_update($sql = "", $message = "") {

		$this->time();
    $this->core->db_call_count++;

		try {

    	$sth = $this->prepare($sql);
    	$var = $sth->execute();

			$return = $sth->rowCount();

			if (not_empty($message))
				$this->core->message->set($var, $message.": ".$sql);

  	}

		catch (PDOException $e) {
			$this->error_info = $e->errorInfo;
			$this->set_rollback();
			$return = false;
			$this->error($e,$sql,$sth);
		}

  	$this->time($sql);
  	return $return;
  }


	public function sql_query($sql, $message = false) {

		if (is_indexed($sql)) {

			foreach ($array = $sql as $sql) {
				$this->sql_query($sql, $message);
			}

			return $message;
		}

		$this->time();
    $this->core->db_call_count++;

		try {
			$sth = $this->prepare($sql);
    	$var = $sth->execute();
    	$message = ($message) ? $this->core->message->set($var, $message." - ".$sql) : true;
  	}

		catch (PDOException $e) {
			$this->error_info = $e->errorInfo;
			$this->set_rollback();
  		$message = $this->error($e,$sql,$sth);
		}
  	$this->time($sql);
  	return $message;
  }


	public function start_transaction() {

  	if (!$this->transaction_active && $this->core->message->set($this->beginTransaction(), "DB - Starting Transaction"))
  		$this->transaction_active = true;

  	return $this->transaction_active;
  }


	public function update($table = "", $values = [], $where = []) {

		if (is_empty($values))
			return true;

		if (any_empty($table, is_array($values)))
			return false;

		if (is_indexed($values)) {

			foreach ($values as $value) {
				$this->update($table, $value, $where);
			}

			return true;

		}

		// Associated array

		foreach ($values as $column => $value) {

			if (is_array($value))
				$value = $this->core->json->encode($value);

			if (!in_list($value,$this->no_string))
				$value = "'".addslashes($value)."'";
				// $value = "'".$value."'";

			$update[] = "`".$column."` = ".$value;

		}

		$update = implode(", ", $update);

		$sql = "UPDATE ".$this->config("prefix").$table." SET $update".$this->where($where).";";

		return $this->sql_update($sql, "UPDATE $table");

	}


	public function where($where = [], $options = []) {

		if (($array = either(key_of("include", $where), key_of("exclude", $where))) && is_array($array) && !is_indexed($array)) {

			$filters = [];

			if (not_empty($include = key_of("include", $where)))
				$filters[] = $this->where($include, array(
					"where" => false,
				));

			if (not_empty($exclude = key_of("exclude", $where)))
				$filters[] = $this->where($exclude, array(
					"where" => false,
					"exclude" => true,
				));

			return "WHERE ".implode(" AND ", $filters);

		}

		if (is_indexed($where)) {

			$filters = [];
			$options = $this->core->array->merge(
				array(
					"where" => false,
					"table" => "",
				),
				$options
			);

			foreach ($where as $w) {
				$filter = $this->where($w, $options);

				if (not_empty($filter))
					$filters[] = $filter;
			}

			if (is_empty($filters))
				return "";

			return "WHERE ".implode(" AND ", $filters);
		}

		$options = $this->core->array->merge(
			array(
				"prefix" => "",
				"where" => true,
				"exclude" => false,
			),
			$options
		);

		if (any_empty($where, is_array($where)))
			return "";

		$filters = [];

		foreach ($where as $column => $values) {

			if ($values === false)
				continue;

			if (is_numeric($column)) {
				$values = ifttt(
					key_of("prefix", $options),
					key_of("prefix", $options).".".$values,
					$values
				);
				$filters[] = $values;
				continue;
			}

			if (is_empty($column) && is_array($values)) {

				$sub_filters = [];

				foreach ($values as $column => $filter) {

					if (not_empty($table = key_of("table", $options))) {
						if (is_array($table))
							$table = current($table);
						$column = $table.".`$column`";
					}

					$sub_filters[] = $this->where(array($column => $filter), array("where"=>false));
				}

				$filters[] = "(".implode(" OR ",$sub_filters).")";

				continue;
			}

			if (not_empty($table = key_of("table", $options))) {
				if (is_array($table)) {
					if (is_numeric(key($table)))
						$table = current($table);
					else
						$table = key($table);
				}
				$column = $table.".`$column`";
			}

			$column = ifttt(
				key_of("prefix", $options),
				key_of("prefix", $options).".".$column,
				$column
			);

			$column = ifttt(
				matches(".",$column),
				$column,
				"`$column`"
			);

			if ($values === NULL)
				$values = "NULL";

			if (is_array($values)) {

				foreach($values as $index => $value)
					if (!in_list($value,$this->no_string))
						$values[$index] = ifttt(
							matches("`",$value),
							$value,
							"'".$value."'"
						);


				$filters[] = "$column ".ifttt(key_of("exclude", $options), "NOT IN", "IN")." (".ifttt($values, implode(",",$values), "NULL").")";

			}

			else if (preg_match_any(["^<= ","^>= ","^BETWEEN"],$values) && matches("'",$values)) {

				$filters[] = "$column ".ifttt(key_of("exclude", $options), "NOT ")."$values";

			}

			else {

				if (!value_fetch($values,$this->no_string) && !matches("`",$values) && !preg_match("/^'(.*)'$/",$values)) {
					$values = preg_replace("/'/","\'",$values);
					$values = "'".$values."'";
				}

				if (value_fetch($values,$this->no_equal))
					$filters[] = "$column ".ifttt(key_of("exclude", $options), "IS NOT", "IS")." $values";

				else if (preg_match("/^'%.+%'$/",$values))
					$filters[] = "$column ".ifttt(key_of("exclude", $options), "NOT LIKE", "LIKE")." $values";

				else
					$filters[] = "$column ".ifttt(key_of("exclude", $options), "!=", "=")." $values";

			}
		}

		$filters = implode(" AND ",$filters);

		if (is_empty($filters))
			return "";

		if (key_of("where", $options) === false)
			return $filters;

		return " WHERE $filters";

	}


	// Private

	private function from ($table = "") {

		if (is_empty($table))
			return $table;

		if (is_string($table))
			return "* FROM ".$this->config("prefix").$table;

		if (is_array($table)) {

			$select = [];
			$from = [];

			foreach ($tables = $table as $table => $options) {

				if (is_numeric($table)) {
					$table = $options;
					$options = "*";
				}

				if (is_array($table = $this->core->json->decode($table))) {
					$table_alias = current($table);
					$table = key($table);
					$from[] = $this->config("prefix").$table. " AS ".$table_alias;
				}

				else {
					$table_alias = false;
					$from[] = $this->config("prefix").$table;
				}


				foreach ((array)$options as $column => $alias) {

					if (is_numeric($column) && is_array($alias)) {

						$on = [];

						foreach ($columns = $alias as $column => $value)
							$on[] = "$table_alias.$column = ".ifttt(matches("`", $value), $value, "'".$value."'");

						$from[] = array_pop($from)." ON ".implode(" AND ", $on);

					}

					else if (is_numeric($column)) {
						$column = $alias;
						if ($table_alias)
							$select[] = $table_alias.".".$column;
						else
							$select[] = $this->config("prefix").$table.".".$column;
					}

					else {
						if ($table_alias)
							$select[] = $table_alias.".".$column." AS ".$alias;
						else
							$select[] = $this->config("prefix").$table.".".$column." AS ".$alias;
					}
				}
			}

			$select = implode(", ", $select);
			$from = implode(" LEFT JOIN ", $from);
			return "$select FROM $from";
		}

		return "";

	}

	private function limit($start = NULL, $count = NULL) {

		if (is_empty($start))
			return "";

		if (is_indexed($start)) {
			$count = key_of(1, $start);
			$start = key_of(0, $start);
		}

		if (is_empty($count))
			return " LIMIT ".intval($start);

		return " LIMIT ".intval($start).", ".intval($count);

	}

	private function order_by($order = []) {

		if (is_empty($order))
			return "";

		if (is_string($order))
			$order = array($order);

		if (!is_array($order))
			return "";

		$order_by = [];

		foreach ($order as $key => $value) {

			if (is_numeric($key)) {
				$key = $value;
				$value = "";
			}

			$order_by[] = ifttt($value, "`$key` $value", "`$key`");
		}

		if (is_empty($order_by))
			return "";

		$order_by = implode(", ",$order_by);


		return " ORDER BY ".$order_by;

	}


	// Uncatergorized

	function time($stop = false) {

		if ($stop === false)
			$this->error_info = [];


		if ($this->calc_time == true) {

			if ($stop) {
				$this->time_end[] =  explode(" ", microtime());
				$this->queries[] = $stop;
				//TODO:$this->error_info = [];
			}

			else {
				$this->time_start[] = explode(" ", microtime());
			}

		}
	}

	function set_dev() {
		$this->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
		$this->dev = true;
		$this->calc_time = true;
	}

	function error($e, $sql, $sth=false) {

		$this->error_info = $sth->errorInfo();

		if (isset($this->dev) && $this->dev) {

			$this->core->message->set(false, "PDO Exception - ".$sql);

			echo $this->core->html->html($this->core->html->body(
				$this->core->html->div(
					[
						$this->core->html->h2("PDO DATABASE EXCEPTION:",["style"=>"font-weight: 300;color: coral;"]),
						$this->core->html->p($sql),
						$this->core->html->p($e->getMessage(),["style"=>"font-weight: 700"]),
						$this->core->html->p($e->getFile()." on line ".$e->getLine()),
						$this->core->html->pre(
							$e->getTraceAsString(),
							["style" => "font-family: ui-monospace, 'Cascadia Code', 'Source Code Pro', Menlo, Consolas, 'DejaVu Sans Mono', monospace; max-width: 100%; overflow: auto; font-size: .9em;"]
						),
					],
					["style" => "padding: 2ch;border: solid 2px coral;max-width: 80ch;word-break: break-all;margin: auto;border-radius: 2ch;"]
				),
				["style" => "display: flex;align-items: center;font-family: system-ui;color: darkslategrey;font-size: 14px;line-height: 130%;"]
			));

			exit();
		}

		$this->set_rollback();
		return false;
	}

	// DEPRECATED

	public function count($table = "", $options = []) {
		return $this->table->rows($table,$options);
	}
	public function table_columns($table) {
		return $this->table->columns($table);
	}
	public function create_table($name = false, $columns = [], $options = []) {
		return $this->table->create($name, $column, $options);
	}
	public function delete_table($name = false) {
		return $this->table->delete($name);
	}
	public function table_exists($name = false) {
		return $this->table->exists($name);
	}

}


class KB_db_table extends KB_class {


	private $core;


	function __construct($core) {

		$this->core = $core;

		$this->config("table_setup", array(
			"ENGINE" => "InnoDB",
			"DEFAULT CHARSET"	=> "utf8",
		));
	}


	public function change($name = "", $columns = []) {

		if (any_empty($name, $columns, is_array($columns)))
			return false;

		$change = [];

		foreach ($columns as $column => $setting) {
			if (any_empty($column, $setting, is_string($setting)))
				continue;

			$change[] = "CHANGE `".trim($column)."` `".trim($column)."` ".$setting;

		}

		if (is_empty($change))
			return false;

		$sql = "ALTER TABLE ".$this->config("prefix").$name." ".implode(", ",$change).";";

		return $this->core->db->query($sql);

	}


	public function columns($table = "", $db = "") {

		if (not_empty($cache = $this->config("db_columns$db")))
			return either(key_of($table,$cache), []);

		if (!$this->exists($table, $db))
			return [];

		$columns = [];

		$sql = "SELECT * FROM `INFORMATION_SCHEMA`.`COLUMNS` WHERE `TABLE_SCHEMA`='".either($db,$this->core->db->name())."';";
		$result = $this->core->db->sql_all($sql);


		foreach($result as $column)
			$columns[$column["TABLE_NAME"]][$column["COLUMN_NAME"]] = $column;

		$this->config("db_columns$db",$columns);

		return either(key_of($table,$columns), []);

	}


	public function create($name = false, $columns = [], $options = []) {

		if (any_empty($name,$columns))
		 	return false;

		$sql = "CREATE TABLE IF NOT EXISTS ".$this->config("prefix").$name." (";

		foreach ($columns as $key => $value){
			$sql_lines[] = "`".$key."` ".$value;
		}

		$sql_lines = $this->core->array->union($sql_lines, $options);

		$sql .= implode(",",$sql_lines);
		$sql .= ")";

		unset($sql_lines);

		foreach ($this->config("table_setup") as $key => $value){
			$sql_lines[] = $key."=".$value;
		}

		$sql .= implode(",",$sql_lines);
		$sql .= ";";

		$this->config("db_columns",[]);
		$this->config("db_tables",[]);
		return $this->core->db->sql_query($sql);

	}


	public function delete($name = "", $db = "") {

		if (is_empty($name))
			return false;

		$name = db_encode($name);
		$sql = "DROP TABLE IF EXISTS ".ifttt($db,$db.".").$this->config("prefix").$name.";";
		$this->config("db_columns",[]);
		$this->config("db_tables",[]);

		return $this->core->db->sql_query($sql);
	}


	public function exists($name = "", $db = "") {

		if (is_empty($name))
			return false;

		if (not_empty($cache = $this->config("db_tables$db")))
			return either(key_of($name,$cache), false);

		$sql = "SELECT * FROM `INFORMATION_SCHEMA`.`TABLES` WHERE `TABLE_SCHEMA`='".either($db,$this->core->db->name())."';";
		$result = $this->core->db->sql_all($sql);

		$tables = [];

		foreach($result as $table)
			$tables[$table["TABLE_NAME"]] = $table;

		$this->config("db_tables$db",$tables);
		return either(key_of($name,$tables), false);

	}


	public function expand($name = "", $columns = []) {

		if (any_empty($name, $columns))
			return false;

		if (is_string($columns))
			$columns = array($columns);

		foreach ($columns as $column => $config) {

			if (is_numeric($column))
				$add[] = "ADD `".trim($config)."` VARCHAR( 255 ) NOT NULL DEFAULT ''";
			else
				$add[] = "ADD `".trim($column)."` $config";
		}

		$sql = "ALTER TABLE ".$this->config("prefix").$name." ".implode(", ",$add).";";
		$this->config("db_columns",[]);
		return $this->core->db->query($sql);

	}


	public function shrink($name = "", $columns = []) {

		if (any_empty($name, $columns))
			return false;

		if (is_string($columns))
			$columns = array($columns);

		foreach ($columns as $column)
			$drop[] = "DROP '".trim($column)."'";

		$sql = "ALTER TABLE ".$this->config("prefix").$name." ".implode(", ",$drop).";";
		$this->config("db_columns",[]);

		return $this->core->db->query($sql);

	}


	public function purge($name = "") {

		if (is_empty($name))
			return false;

		$sql = "DELETE FROM TABLE ".$this->config("prefix").$name." WHERE 1;";
		return $this->core->db->query($sql);

	}


	public function sort($table = "", $sort = "", $sort_by = "", $where = []) {

		if (any_empty($table, $sort, $sort_by))
			return false;

		$sql = "
			SET @c := 0;
			INSERT INTO ".$this->config("prefix").$table." (ID, $sort) SELECT ID, c FROM (
				SELECT ID, @c:=@c+1 AS c FROM ".$this->config("prefix").$table.$this->core->db->where($where)." ORDER BY `$sort_by`
			) AS sub ON DUPLICATE KEY UPDATE $sort=VALUES($sort);
		";

		return $this->core->db->query($sql);
	}


	// Deprecated

	public function rows($table = "", $where = []) {
		deprecated('Use $core-db->rows() instead of $core->db->table->rows()');
		return $this->core->db->rows($table, $where);

	}

}