<?php


class KB_filesystem extends KB_class_extended {

	public
		$file,
		$folder,
		$scope = "inner";


	protected
		$core,
		$fs;


	function __construct ($core) {
		$this->fs = $this;
		$this->core = $core;
		$this->file = new KB_file($core);
		$this->file->fs = $this;
		$this->folder = new KB_folder($core);
		$this->folder->fs = $this;
	}


	// Public


	public function base_dir ($set = NULL) {

		if (is_null($set))
			return either($this->config("base_dir"), "");

		return $this->set_base_dir($set);


	}


	public function copy ($source = "", $target = "", $options = []) {

		if (is_indexed($source)) {
			$return = array();
			foreach ($source as $source_entry)
				$return[] = $this->copy($source_entry, $target);
			return $return;
		}

		if (are_equal($source, $target))
			return true;

		if (is_file($this->get_path($source)))
			return $this->fs->file->copy($source, $target);

		if (is_dir($this->get_path($source)))
			return $this->fs->folder->copy($source, $target);

		return false;

	}


	public function delete ($file = "") {

		if (is_empty($file))
			return true;

		clearstatcache();

		if (is_indexed($file)) {
			foreach ($file as $file_entry)
				$return[] = $this->delete($file_entry);
			return $return;
		}

		if (!file_exists($this->get_path($file)))
			return true;

		if (is_file($this->get_path($file)))
			return $this->file->delete($file);

		if (is_dir($this->get_path($file)))
			return $this->folder->delete($file);

		return true;
	}


	public function dir ($set = NULL) {

		return $this->base_dir($set);

	}


	public function download ($filename = "") {

		if (any_empty($filename, is_string($filename)))
			return;

		$file = $this->get_path($filename);
		$this->send_file($filename, stat($file), $this->file_type($filename), 0,$filename);
	}


	public function exists ($path) {

		if ($this->file->exists($path))
			return true;

		if ($this->folder->exists($path))
			return true;

		return false;

	}


	public function get_path ($path = "", $file = "") {

		if (is_indexed($paths = $path) && is_empty($file)) {

			foreach ($paths as $key => $path)
				$paths[$key] = $this->get_path($path);

			return $paths;

		}

		if (!is_string($path))
			return $this->get_path("", $file);

		if (not_empty($file)) {
			$path = preg_replace("/\/$/","",$path);
			$path = $path."/".$file;
		}

		// Remove trailing slash and spaces
		$path = trim($path, " \t\n\r");
		// $path = preg_replace("/\/$/", "", $path);

		// Check if path is valid
		// Disable error_reporting since PHP cannot catch this
		$error_level = error_reporting(0);

		if (is_dir($path) || is_file(preg_replace("/\/$/", "", $path))) {
			error_reporting($error_level);
			return $path;
		}

		if (!matches($this->core->path("root"),$path)) {

			if (!preg_match(preg_mod("^".$this->dir()),$path))
				$path = $this->dir().$path;

			if (!matches($this->core->path("root"),$path))
				$path = $this->core->path("root").$path;

		}

		// add trailing slash
		if (is_dir($path))
			$path = preg_replace("/([^\/])$/", "$1/", $path);

		// remove trailing slash
		if (is_file($file = preg_replace("/\/$/", "", $path)))
			$path = $file;

		error_reporting($error_level);

		return $path;
	}


	public function move ($source = "", $target = "") {

		if (are_equal($source, $target))
			return true;

		if (!$this->copy($source, $target))
			return false;

		return $this->delete($source);
	}


	public function path () {

		$path = "";

		if (not_empty($args = func_get_args()))
			foreach($args as $folder) {

				$folder = preg_replace("/[\/]{2,}/","/",$folder);

				if (!matches($this->core->path("root"), $folder))
					$folder = preg_replace("/^\//","",$folder);

				$folder = preg_replace("/\/$/","",$folder);

				if(not_empty($folder))
					$path .= "$folder/";
			}
		return $path;
	}


	public function read ($path = "", $options = []) {

		if ((count(func_get_args()) === 1) && is_array($path)) {

			if (is_indexed($path, true)) {

				$return = [];

				foreach ($paths = $path as $path)
					$return = $this->core->array->union($return, $this->read($path));

				return $return;

			}

			else {

				$options = $path;
				$path = "";

			}

		}


		$options = array_merge(array(
			"recursive" => false,
			"type" => "file",
			"full_path" => false,
			"path" => false,
			"filter" => false,
			"filetype" => false,
			"filetype_exclude" => false,
			"return" => "array",
		),$options);

		if (is_file($this->get_path($path)))
			return $this->file->read($path);

		$array = array();

		if(fetch("type",$options) == "folder"){

			$folders = $this->scan_for_folders($path);

			if(fetch("filter",$options))
				$folders = array_find(fetch("filter",$options),$folders);

			if(fetch("recursive",$options)){

				foreach($folders as $key => $folder)
						$new_array[$folder] = $this->read($path.$folder."/", $options);

				if(!empty($new_array)){
					$folders = $new_array;
					unset($new_array);
				}
			}

			$array = $array + $folders;
		}

		if(fetch("type",$options) == "file"){


			$files = $this->scan_for_files($path);

			if(fetch("filter",$options))
				$files = array_find(fetch("filter",$options),$files);

			if ($filetype = key_of("filetype",$options)) {
				$files = array_find(".".$filetype,$files,"");
			}

			if (key_of("full_path", $options))
				foreach ($files as $key => $value)
					$files[$key] = $this->get_path($path."/".$value);

			else if (key_of("path", $options))
				foreach ($files as $key => $value)
					$files[$key] = $path."/".$value;



			$array = $array + $files;
		}

		return $array;
	}


	public function read_php($path, $local = [], $return = []){

		// Make $core, $kit available

		$core = $this->core;

		if (isset($this->core->kit->current))
			$kit = $this->core->kit->current;

		if (not_empty($local) && is_array($local))
			foreach ($local as $name => &$value)
				$$name = &$value;

		// Prevent bug that overrides last set variable
		unset($value);

		if (is_file($this->get_path($path))) {

			ob_start();

			try {
				include $this->get_path($path);
			}
			catch (Throwable $e){
				return $e;
			}
			catch (Exception $e){
				return $e;
			}

			if (is_empty($return))
				return ob_get_clean();

			ob_end_flush();

			if (is_string($return) && isset($$return))
				return $$return;

			if (is_array($return)) {

				$array = array();

				foreach ($return as $variable) {
					if (is_string($variable) && isset($$variable))
						$array[$variable] = $$variable;
				}

				return $array;
			}

			return true;
		}
		return false;
	}


	public function rename ($old = "", $new = "") {

		if (are_equal($old, $new))
			return true;

		if (any_empty($old, $new))
			return false;

		if (is_file($this->get_path($old)))
			return $this->rename_file($old, $new);

		if (is_dir($this->get_path($old)))
			return $this->rename_folder($old, $new);

		return false;

	}


	public function search ($file = "", $path = "") {

		if (is_empty($file))
			return false;

		$return = [];

		if (is_indexed($file)) {
			foreach ($file as $f) {
				$return = $this->core->array->merge(
					$return,
					$this->search($f, $path)
				);
			}
			return $return;
		}

		$path = $this->core->folder->part($file).$path;

		$file = $this->core->file->name($file);

		$files = $this->read($path);

		if (value_fetch($file,$files))
			$return[] = ifttt($path, $path."/".$file, $file);

		if (not_empty($folders = $this->scan($path))) {

			foreach ($folders as $folder) {

				if (not_empty($files = $this->search($file, $path."/".$folder)))
					$return = $this->core->array->merge($return,$files);
			}
		}

		return $return;
	}


	public function size ($dir = "")	{

			$dir = $this->get_path($dir);

			$size = 0;
			$entries = glob(rtrim($dir, '/').'/*', GLOB_NOSORT);

			if(!empty($entries))
				foreach ($entries as $each)
					$size += is_file($each) ? filesize($each) : $this->size($each);

			return $size;
	}


	public function timestamp ($path = "") {

		if (is_file($this->get_path($path)))
			return $this->file->age($path,array("age"=>"modified"));

		if (is_dir($this->get_path($path))) {

			// add trailing slash
			$path = preg_replace("-([^\/])$-","$1/",$path);

			$age = array();

			foreach ($this->folder->read($path) as $sub_path)
				$age[] = $this->timestamp($path.$sub_path);

			rsort($age);
			return current($age);
		}

		return 0;
	}


	public function touch ($source = "", $filemtime = "") {

		if (is_empty($source))
			return false;

		if (!$this->exists($source))
			return false;

		if (is_empty($filemtime))
			return touch($this->get_path($source));

		return touch($this->get_path($source), $filemtime);

	}






	// Deprecated

	public function age ($file="", $options = []) {
		deprecated();
		return $this->file->age($file,$options);
	}

	public function chmod($folder, $mode){

		deprecated();

		return;

		passthru("chmod ".$mode." ".DROOT.$folder, $err);

		$this->core->message->set($err,'chmod changed via PHP','chmod can’t be changed via PHP, trying FTP');

		if ($err) {
			$ftp = $this->core->ftp_o();
			$ftp_con_id = $ftp->chmod_open();
			$this->core->message->set($ftp->chmod_file($ftp_con_id, $mode, $folder), 'chmod changed via FTP', 'chmod can’t be changed via FTP');
			$ftp->chmod_close($ftp_con_id);
		}
	}

	public function create_document ($filename,$folder="",$content="") {
		deprecated();
		return $this->create_file($filename,$folder,$content);
	}

	public function create_file ($filename,$folder="",$content="") {
		deprecated();
		return $this->file->create($filename,$folder,$content);
	}

	public function create_folder ($path="") {

		deprecated();

		if ($this->exist_folder($path))
			return true;

		else
			return $this->core->message->set(mkdir($this->get_path($path), 0755, true), "Create folder ".$this->get_path($path));

	}

	public function delete_file ($file = "") {

		deprecated();
		// deprecated('Will become private. Use $core->filesystem->delete instead of $core->filesystem->delete_file');

		if(is_empty($file))
			return true;

		clearstatcache();

		if(!file_exists($this->get_path($file)))
			return true;

		return unlink($this->get_path($file));

	}

	public function delete_folder ($folder = ""){

		deprecated();

		if(is_empty($folder))
			return true;

		if(!$this->exist_folder($folder))
			return true;

		foreach($this->scan_dir($folder) as $entry){
			self::delete($folder.'/'.$entry);
		}

		return rmdir($this->get_path($folder));
	}

	public function exist_folder ($path) {
		deprecated();
		return $this->folder->exists($path);
	}

	public function read_file ($path) {

		deprecated();
		return $this->file->read($path);
	}

	public function rename_folder ($path_old, $path_new) {

		deprecated();
		if(
			$this->core->message->set(
				file_exists($this->get_path($path_old)), "old folder_exist ".$this->get_path($path_old)
			) &&
			$this->core->message->set(
				!file_exists($this->get_path($path_new)),"new folder_NOT_exist :".$this->get_path($path_new)
			)
		){
			// Create directory tree
			$path_new = preg_replace("-/$-","",$path_new);

			if(preg_match("-/-",$path_new)){
				$path = "";
				$folders = preg_split("-/-",$path_new);
				$folders = array_slice($folders,0,-1);
				foreach($folders as $folder){
					$path = $path.$folder."/";
					$this->create_folder($path);
				}
			}
			// Move and return
			return $this->core->message->set(rename($this->get_path($path_old), $this->get_path($path_new)), "Rename: ".$path_old."->".$path_new);
		}
		return false;
	}

	public function scan_dir($path, $options = array()) {

		deprecated();

		$options = $this->core->array->union($options, [
			"name" => key_of("filter", $options),
			"extension" => key_of("filetype", $options),
			"extension_exclude" => key_of("filetype_exclude", $options),
		]);

		return $this->folder->read($path, $options);

	}

	public function scan_dir_ext ($path = "", $options = []) {

		deprecated();

		$options = $this->core->array->union($options, [
			"stats" => true,
			"name" => key_of("filter", $options),
			"extension" => key_of("filetype", $options),
			"extension_exclude" => key_of("filetype_exclude", $options),
		]);

		return $this->folder->read($path, $options);

	}

	public function write ($filename, $content = ""){

		deprecated();
		$this->file->delete($filename);
		return $this->file->create($filename,"",$content);
	}

	// Deprecated: Will be private

	function rename_file($path_old, $path_new) {

		if (are_equal($path_old, $path_new))
			return true;

		if ($this->exist_file($path_old)) {

			$this->file->delete($path_new);

			// Create directory tree
			if(preg_match("-/-",$path_new)){
				$path = "";
				$folders = preg_split("-/-",$path_new);
				$folders = array_slice($folders,0,-1);
				foreach($folders as $folder){
					$path = $path.$folder."/";
					$this->create_folder($path);
				}
			}
			return $this->core->message->set(rename($this->get_path($path_old), $this->get_path($path_new)), "Rename: ".$path_old."->".$path_new);
		}
		return false;
	}



	// Private

	protected function send_file($path, $stat, $ftype, $cache, $filename=false) {
		if($this->exist_file($path)){

			header("Accept-Ranges: bytes");
			if($cache == 1){
				header('Etag: "'.$stat['mtime'].'"');
				header("Last-Modified: ".date("D, d M Y H:i:s", intval($stat['mtime']))." GMT");
				header("Expires: ".date( 'D, d M Y H:i:s', time()+604800 )." GMT");
				header("Cache-Control:	max-age=604800");
			}
			header("Content-Length: ".$stat['size']);
			header("Content-Type: ".$ftype);
			if($filename) header('Content-Disposition: attachment; filename="'.$filename.'"');
			//echo $this->file_string($path);
			ob_end_clean();
			set_time_limit(0);
			/*
			while (ob_get_level() > 0) {
					ob_end_flush();
			}
			set_time_limit(0);
			*/
			// header("X-Sendfile: ".$this->get_path($path));
			readfile($this->get_path($path));
			exit(1);
		}
	}

	private function scan_for_folders ($path="") {

		if(is_dir($this->get_path($path))){

			$dir = scandir($this->get_path($path));

			foreach($dir as $key => $value){

				if($value == ".")
					unset($dir[$key]);

				else if($value == "..")
					unset($dir[$key]);

				else if(!is_dir($this->get_path($path."/".$value)))
					unset($dir[$key]);

			}

			sort($dir);
			return $dir;

		}else{

			return array();

		}

	}

	private function scan_for_files ($path="") {

		if(is_dir($this->get_path($path))){

			//Path should end on slash
			$path = preg_replace("-([^/])$-","$1/",$path);

			$dir = scandir($this->get_path($path));

			foreach($dir as $key => $value){

				if($value == ".." || !is_file($this->get_path($path).$value))
					unset($dir[$key]);
			}

			sort($dir);
			return $dir;

		}else{

			return array();

		}

	}

	protected function set_base_dir($path = "") {

		if (!is_string($path))
			return $this->base_dir();

		// Add trailing slash
		$path = preg_replace("-([^\/])$-","$1/",$path);
		$this->config("base_dir", $path);
		return $this->base_dir();

	}


	// Uncatogrized methods, needs cleanup

	public function scan ($path="") {

		if($this->scope == "outer"){
			if(is_dir($path)) $dir = scandir($path);
		}else{
			if(is_dir($this->get_path($path))) $dir = scandir($this->get_path($path));
		}

		if(isset($dir)){

			foreach($dir as $key => $value){

				if($value == ".")	unset($dir[$key]);
				else if($value == "..") unset($dir[$key]);
				else if( $this->scope == "outer" && !is_dir($path."/".$value))	unset($dir[$key]);
				else if( $this->scope == "inner" && !is_dir($this->get_path($path."/".$value)))	unset($dir[$key]);

			}

			sort($dir);
			return $dir;

		}else{
			return false;
		}
	}

	public function read_path($path = "", $options = array()){

		$array = $this->read($path, $options);

		$return = array();

		while(!empty($array)){

			$return[] = $this->return_path($array);
		}

		return $return;

	}

	private function return_path(&$array){

		$return = "";

		$i = 0;

		${"current".$i} =& $array;

		while(is_array(${"current".$i})){
			$return .= (key(${"current".$i}) == "") ? "" : key(${"current".$i})."/";
			${"current".($i+1)} =& ${"current".$i}[key(${"current".$i})];
			$i++;
		}
		unset(${"current".($i-2)}[key(${"current".($i-2)})]);

		return $return;

	}

	function chmod_need($ordner_php,$ordner, $mode){

		if(isset($_ENV["FCGI_ROLE"]))
			return true;

		$this->chmod($ordner_php,$ordner, $mode);
	}

	public function copy_folder($file_old, $file_new,$options=array()){

		deprecated();

		if ($this->exist_folder($file_old)) {

			if ($this->exist_folder($file_new)) {

				if (!in_array("merge",$options)) {
					$this->delete_folder($file_new);
					$this->create_folder($file_new);
				}
			}

			else{
				$this->create_folder($file_new);
			}


			$dir = opendir($this->get_path($file_old));

			while(false !== ( $file = readdir($dir)) ) {

				if (( $file != '.' ) && ( $file != '..' )) {
					if ( is_dir($this->get_path($file_old).'/'. $file) ) {
						$this->copy_folder($file_old.'/'.$file, $file_new.'/'.$file);
					}
					else {
						if(!in_array("keep",$options) || (in_array("keep",$options) && !is_file($this->get_path($file_new).'/'.$file))){
							if(
								copy($this->get_path($file_old).'/'.$file, $this->get_path($file_new).'/'.$file)
							){
								touch(
									$this->get_path($file_new).'/'.$file,
									filemtime($this->get_path($file_old).'/'.$file)
								);
							}
						}
					}
				}
			}
			closedir($dir);
			// return true;

			// $this->core->message->set($this->recurse_copy($this->get_path($file_old), $this->get_path($file_new), $options), "Copy Folder: ".$file_old."->".$file_new);

			if(in_array("move",$options))
			$this->core->message->set($this->delete_folder($file_old), "Delete source:".$file_old);

			return true;
		}
		return false;
	}

	function copy_file($source, $target){

		if($source == $target)
			return true;

		if(!$this->exist_file($source))
			return false;

		// Case copy file into folder
		if (is_dir($this->get_path($target)) || preg_match("/(.+)\/$/",$target)) {

			// $this->create_folder($target);
			preg_match("/([^\/]+)$/",$source,$matches);
			$filename = $matches[0];
			$target = $target.$filename;
			return $this->copy_file($source, $target);

		}

		// Extract path from target to create folders if neccessarry
		if (matches("/",$target)) {

			$path = explode("/",$target);
			$folders = array_slice($path,0,-1);


			if(not_empty($folders)) {
				$folders = implode("/",$folders);
				$this->create_folder($folders);
			}

		}

		if (is_file($this->get_path($target)))
			$this->delete_file($target);

		if (copy($this->get_path($source), $this->get_path($target))) {

			touch($this->get_path($target),filemtime($this->get_path($source)));
			return true;
		}

		return false;
	}

	function open_file($path, $mode="r") {
		$handle = fopen($this->get_path($path),$mode);
		$this->handle = $handle;
		return $handle;
	}

	function close_file($handle=false) {
		if(!$handle)
		{
			fclose($this->handle);
		}
		else{ fclose($handle); }
	}

	function file_array($path) {

		$lines = file($this->get_path($path), FILE_IGNORE_NEW_LINES);
		return $lines;
	}

	function file_string($path) {

		$file = $this->open_file($path);
		$data='';
		while(!feof($file))
				$data.=fread($file,1024);
		rewind($this->handle);
		//$this->close_file();

		return $data;
	}

	public function exist_file($file) {
		return file_exists($this->get_path($file));
	}

	function file_type($file){
				$mime_types = array(

						'txt'	=> 'text/plain',
						'htm'	=> 'text/html',
						'html'	=> 'text/html',
						'php'	=> 'text/html',
						'css'	=> 'text/css',
						'js'	=> 'application/javascript',
						'json'	=> 'application/json',
						'xml'	=> 'application/xml',
						'swf'	=> 'application/x-shockwave-flash',
						'flv'	=> 'video/x-flv',
			 		'csv'	=> 'text/comma-separated-values; charset=utf-8',

						// images
						'png' => 'image/png',
						'jpe' => 'image/jpeg',
						'jpeg' => 'image/jpeg',
						'jpg' => 'image/jpeg',
						'gif' => 'image/gif',
						'bmp' => 'image/bmp',
						'ico' => 'image/vnd.microsoft.icon',
						'tiff' => 'image/tiff',
						'tif' => 'image/tiff',
						'svg' => 'image/svg+xml',
						'svgz' => 'image/svg+xml',

						// archives
						'zip' => 'application/zip',
						'rar' => 'application/x-rar-compressed',
						'exe' => 'application/x-msdownload',
						'msi' => 'application/x-msdownload',
						'cab' => 'application/vnd.ms-cab-compressed',

						// audio/video
						'mp3' => 'audio/mpeg',
						'qt' => 'video/quicktime',
						'mov' => 'video/quicktime',

						// adobe
						'pdf' => 'application/pdf',
						'psd' => 'image/vnd.adobe.photoshop',
						'ai' => 'application/postscript',
						'eps' => 'application/postscript',
						'ps' => 'application/postscript',

						// ms office
						'doc' => 'application/msword',
						'rtf' => 'application/rtf',
						'xls' => 'application/vnd.ms-excel',
						'ppt' => 'application/vnd.ms-powerpoint',

						// open office
						'odt' => 'application/vnd.oasis.opendocument.text',
						'ods' => 'application/vnd.oasis.opendocument.spreadsheet',
				);
				$ext = strtolower($this->core->array->pop(explode('.',$file)));

				if (array_key_exists($ext, $mime_types)) {
						return $mime_types[$ext];
				}
				elseif (function_exists('finfo_open')) {
						$finfo = finfo_open(FILEINFO_MIME);
						$mimetype = finfo_file($finfo, $file);
						finfo_close($finfo);
						return $mimetype;
				}
				else {
						return 'application/octet-stream';
				}
	 }

	function zip_dir($path, $zip_file) {

		$zip = new KB_zip;

		if
		($this->core->message->set

				(
					(
						($res = $zip->open($this->get_path("", $zip_file),	ZIPARCHIVE::CREATE | ZIPARCHIVE::OVERWRITE)) === TRUE
						?
							true
						:
							false
					),

					'ZIP open: '.$this->get_path("", $zip_file).' - '.$zip->error_code($res)
				)

		)
		{
			if(is_array($path))
			{
				foreach($path AS $p)
				{
					$zip->addFolderToZip($p,$this->get_path());
				}
			}
			else
			{
				$zip->addFolderToZip($path,$this->get_path());
			}
			$zip->close();


			return true;
		}
		return false;
	}

	function unzip_file($file, $path="") {

		$zip = new KB_zip;

		if ($this->core->message->set( ( ($res = $zip->open($this->get_path($file))) === TRUE) ? true:false ,'ZIP open: '.$this->get_path($file).' - '.$zip->error_code($res)) && $this->exist_folder($path))
		{
			$this->core->message->set($zip->extractTo($this->get_path($path)),'ZIP extract '.$file.' to '.$path);
			$zip->close();
			return true;
		}
		return false;
	}

	function fopen($filename, $mode) {
			$file = @fopen($filename, $mode);
			$bom = fread($file, 3);
			if ($bom != "\xEF\xBB\xBF"){
					rewind($file);
		}
			return $file;
	}

	function fwrite ($handle, $value="") {
		//fwrite($handle, "\xEF\xBB\xBF".$value);
		//if(is_string($value)){echo "<pre>STRING</pre>";}
		fwrite($handle, $value);
	}

	function get_remote_file($url,$save_path=false) {

		if(!empty($url)) {

			$remote="";
			$ch = curl_init($url);
			curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
			curl_setopt($ch, CURLOPT_HEADER, false);
			$this->core->msg->set($remote = curl_exec($ch),'getting file '.$url);
			$http = curl_getinfo($ch, CURLINFO_HTTP_CODE);
			($http != '200') ? $this->core->msg->set( false,'HTTP CODE: '.$http):true;
			curl_close($ch);

			if(!empty($save_path))
			{
				$this->file->create($save_path, "", $remote);
			}

			return $remote;
		}
	}

	function version_by_date() {

		$result = array_merge($this->get_last_mod_date(DROOT.CORE_FOLDER),$this->get_last_mod_date(DROOT.CORE_FOLDER."library"));
		sort($result);
		return array_pop($result);
	}

	function get_last_mod_date($dirName) {

		$dir = dir($dirName);
		while($file = $dir->read())
		{
				if($file != '.' && $file != '..' && $file != '.svn' && $file != '_notes') {
					$time = filemtime($dirName."/".$file);
		(!empty($time)) ? $times[] = $time : false;
				}

		}
		return $times;
	}

	function export_file($content="",$filename=false){

			$this->set_base_dir(KB_FOLDER);
			if($this->exist_folder("export_temp", false) === false)	$this->create_folder("export_temp");
			$this->create_file($filename,"export_temp",$content);
			$file = DROOT.KB_FOLDER."export_temp/".$filename;
			$this->send_file("export_temp/".$filename, stat($file), $this->file_type($filename), 0,$filename);
			$this->delete_folder("export_temp");
			exit();
	}




	// Alias

	public function cd ($path) {
		return $this->base_dir($path);
	}

	public function cp ($source = "", $target = "") {
		return $this->copy($source, $target);
	}

	public function find($file = "", $path = "") {
		return $this->search($file, $path);
	}

	public function mv ($path = "", $target = "") {
		return $this->move($path, $target);
	}

	public function rm ($path = "") {
		return $this->delete($path);
	}



}


class KB_file extends KB_filesystem {


	protected $core;

	public $fs;


	function __construct($core) {
		$this->core = $core;
	}


	public function age ($file = "", $options = []) {

		if (any_empty($file, $this->exists($file)))
			return false;

		if ($options === true)
			$options = ["set" => date("U")];

		else if (is_numeric($options))
			$options = ["set" => $options];

		else if (!is_array($options))
			$options = [];

		$options = array_merge(
			array(
				"set"=>false,
				"age"=>"max",
			),$options
		);

		if ($set = key_of("set",$options))
			touch($this->get_path($file),$set);

		$age[0] = filemtime($this->get_path($file));
		$age[1] = filectime($this->get_path($file));

		switch (key_of("age", $options)) {

			case "max":
				rsort($age);
				return current($age);

			case "modified":
				return $age[0];

			case "created":
				return $age[1];

			default:
				return 0;
		}
	}


	public function base_dir ($set = NULL) {

		if (is_null($set))
			return $this->fs->config("base_dir");

		return $this->fs->set_base_dir($set);


	}


	public function basename ($file = "") {

		$file = $this->name($file);
		$ext = $this->extension($file);
		$file = preg_replace("/\.$ext$/","", $file);

		return $file;

	}


	public function chmod($file = "", $mode = 0644) {

		if (any_empty($file, $this->exists($file), is_numeric($mode)))
			return false;

		return chmod($this->get_path($file), $mode);

	}


	public function copy ($source = "", $target = "", $options = []) {

		if (is_indexed($sources = $source)) {

			$return = [];

			foreach ($sources as $source)
				$return[] = $this->copy($source, $target, $options);

			return $return;

		}

		$options = $this->core->array->merge([
			"overwrite" => true,
		],$options);

		if (is_empty($source))
			return false;

		if (!$this->exists($source))
			return false;

		$file = $this->name($target);
		$folder = $this->fs->folder->part($target);

		if (is_empty($file))
			$file = $this->name($source);

		$target = $folder.$file;

		if ($this->exists($target) && !$options["overwrite"])
			return true;

		$this->delete($target);
		$this->fs->folder->create($folder);

		if (copy($this->get_path($source), $this->get_path($target))) {
			touch($this->get_path($target),filemtime($this->get_path($source)));
			return true;
		}

		return false;

	}


	public function create ($filename = "", $folder = "", $content = "") {

		clearstatcache();

		$folder = $this->fs->folder->path($folder, $this->fs->folder->part($filename));

		$filename = $this->name($filename);
		$this->fs->folder->create($folder);

		if (($file = fopen($this->get_path($folder,$filename),"w")) === false)
			return false;

		if (fwrite($file, $content) !== false && fclose($file) !== false)
			return true;

		return false;
	}


	public function delete ($file = "") {

		if (is_empty($file))
			return true;

		if (is_indexed($file)) {
			foreach ($file as $file_entry)
				$return[] = $this->delete($file_entry);
			return $return;
		}

		clearstatcache();

		if (!$this->exists($file))
			return true;

		return unlink($this->get_path($file));

		clearstatcache();

	}


	public function dir ($set = NULL) {

		return $this->base_dir($set);

	}


	public function download ($path = false) {

		if (is_empty($path))
			return false;

		$file = $this->get_path($path);
		$filename = $this->name($path);

		if (!is_file($file) || !file_exists($file))
			return;

		$this->send_file($path, stat($file), $this->file_type($filename), 0,$filename);

	}


	public function exists ($path) {

		if (file_exists($this->get_path($path)) && is_file($this->get_path($path)))
			return true;

		return false;

	}


	public function extension ($filename = "") {

		if (is_empty($filename))
			return false;

		$filename = $this->name($filename);

		if (!matches(".",$filename))
			return false;

		$extension = explode(".",$filename);
		$extension = array_pop($extension);
		$extension = strtolower($extension);

		return $extension;

	}


	public function move ($source = "", $target = "", $options = []) {

		if (are_equal($source, $target))
			return $target;

		if (are_equal($this->get_path($source), $this->get_path($target)))
			return $target;

		if ($this->copy($source, $target, $options))
			$this->delete($source);

		else
			return $source;

		return $target;

	}


	public function name ($file = "", $rename = "") {

		if (is_empty($file))
			return $file;

		if (is_indexed($files = $file)) {
			foreach ($files as $key => $file) {
				$files[$key] = $this->name($file, $rename);
			}
			return $files;
		}

		if (not_empty($rename))
			return $this->name($this->move($file, $rename));


		$file_parts = preg_split("/\?/",$file);

		$path = key_of(0,$file_parts);
		$path = preg_split("/\//",$path);

		$basename = current(array_slice($path, -1));

		if (matches(".",$basename))
			return $basename;

		return "";
	}


	public function read ($file = "", $options = []) {

		if (preg_match("-^//-", $file)) {

			if (are_equal("80",key_of("SERVER_PORT",$_SERVER)))
				return $this->read("http:".$file);

			else
				return $this->read("https:".$file);
		}

		if (preg_match("/^http/", $file)) {

			set_error_handler("error_handler");

			try {
				$file_content = file_get_contents($file);
			}

			catch (Exception $e) {
				$file_content = false;
			}

			restore_error_handler();
			return $file_content;
		}

		if (is_file($file = $this->get_path($file))){
			ob_start();
			readfile($file);
			return ob_get_clean();
		}
		return false;

	}


	public function size ($file = "")	{

		if (is_empty($file))
			return NULL;

		$path = $this->get_path($file);

		if (!is_file($path))
			return NULL;

		return filesize($path);

	}


	public function type ($filename = "") {

		if (is_empty($filename))
			return NULL;

		$mime_types = array(

			'txt'	=> 'text/plain; charset=utf-8',
			'htm'	=> 'text/html; charset=utf-8',
			'html'	=> 'text/html; charset=utf-8',
			'php'	=> 'text/html; charset=utf-8',
			'css'	=> 'text/css',
			'js'	=> 'application/javascript',
			'json'	=> 'application/json',
			'xml'	=> 'application/xml',
			'swf'	=> 'application/x-shockwave-flash',
			'flv'	=> 'video/x-flv',
			'csv'	=> 'text/comma-separated-values',

			// images
			'png' => 'image/png',
			'jpe' => 'image/jpeg',
			'jpeg' => 'image/jpeg',
			'jpg' => 'image/jpeg',
			'gif' => 'image/gif',
			'bmp' => 'image/bmp',
			'ico' => 'image/vnd.microsoft.icon',
			'tiff' => 'image/tiff',
			'tif' => 'image/tiff',
			'svg' => 'image/svg+xml',
			'svgz' => 'image/svg+xml',

			// archives
			'zip' => 'application/zip',
			'rar' => 'application/x-rar-compressed',
			'exe' => 'application/x-msdownload',
			'msi' => 'application/x-msdownload',
			'cab' => 'application/vnd.ms-cab-compressed',

			// audio/video
			'mp3' => 'audio/mpeg',
			'wav' => 'audio/wav',
			'qt' => 'video/quicktime',
			'mov' => 'video/quicktime',
			'mp4' => 'video/mp4',

			// adobe
			'pdf' => 'application/pdf',
			'psd' => 'image/vnd.adobe.photoshop',
			'ai' => 'application/postscript',
			'eps' => 'application/postscript',
			'ps' => 'application/postscript',

			// ms office
			'doc' => 'application/msword',
			'rtf' => 'application/rtf',
			'xls' => 'application/vnd.ms-excel',
			'ppt' => 'application/vnd.ms-powerpoint',

			// open office
			'odt' => 'application/vnd.oasis.opendocument.text',
			'ods' => 'application/vnd.oasis.opendocument.spreadsheet',
		);

		$ext = $this->extension($filename);

		if ($mime_type = key_of($ext,$mime_types)) {
				return $mime_type;
		}

		else if (function_exists('finfo_open') && is_file($filename)) {
				$finfo = finfo_open(FILEINFO_MIME);
				$mimetype = finfo_file($finfo, $filename);
				finfo_close($finfo);
				return $mimetype;
		}
		else {
				return 'application/octet-stream';
		}
	}


	public function unzip ($file = "", $path = "") {

		if (any_empty($file, is_string($file), is_string($path)))
			return false;

		if (is_empty($path))
			$path = $this->fs->folder->part($file);

		$zip = $this->core->zip();
		$zip->dir($this->dir());
		return $zip->extract($file, $path);

		/*
		if ($this->core->message->set( ( ($res = $zip->open($this->get_path($file))) === TRUE) ? true:false ,'ZIP open: '.$this->get_path($file).' - '.$zip->error_code($res)) && $this->folder->exists($path))
		{
			$this->core->message->set($zip->extractTo($this->get_path($path)),'ZIP extract '.$file.' to '.$path);
			$zip->close();
			return true;
		}
		return false;
		*/
	}


	public function write ($filename, $content = "") {

		$this->delete($filename);
		return $this->create($filename, "", $content);
	}
}


class KB_folder extends KB_filesystem {


	protected $core;

	public $fs;

	function __construct($core) {
		$this->core = $core;
	}


	public function base_dir ($set = NULL) {

		if (is_null($set))
			return $this->fs->config("base_dir");

		return $this->fs->set_base_dir($set);


	}


	public function copy ($source = "", $target = "", $options = []) {

		$options = $this->core->array->merge([
			"merge" => true,
			"overwrite" => true,
		],$options);

		if (is_indexed($source)) {

			$return = [];

			foreach ($source as $folder) {
				// $return[] = $this->copy($folder, $this->path($target, $this->name($folder)), $options);
				$return[] = $this->copy($folder, $target, $options);
			}

			return $return;
		}

		if (!$this->exists($source))
			return false;

		$target = $this->path($target, $this->name($source, true));

		if (!$options["merge"])
			$this->delete($target);

		$this->create($target);

		foreach ($this->read($source, ["folders"=>false]) as $file)
			$this->fs->file->copy($this->path($source).$file, $target, keys_of(["overwrite"], $options));

		foreach ($this->read($source, ["files"=>false]) as $folder)
			$this->copy($this->path($source).$folder, $target, $options);

		return true;
	}


	public function create ($path = "") {

		if (any_empty(is_string($path)))
			return;

		if ($this->exists($path))
			return true;

		else
			return $this->core->message->set(mkdir($this->get_path($path), 0755, true), "Create folder ".$this->get_path($path));

	}


	public function delete ($folder = "") {

		if (is_empty($folder))
			return true;

		if (is_indexed($folder)) {
			foreach ($folder as $folder_entry)
				$return[] = self::delete($folder_entry);
			return $return;
		}

		if (!$this->exists($folder))
			return true;

		foreach ($this->read($folder,["folders"=>false]) as $file)
			$this->fs->file->delete("$folder/$file");

		foreach ($this->read($folder,["files"=>false]) as $sub_folder)
			$this->delete("$folder/$sub_folder");

		return rmdir($this->get_path($folder));

	}


	public function dir ($set = NULL) {

		return $this->base_dir($set);

	}


	public function exists ($path) {

		if (file_exists($this->get_path($path)) && is_dir($this->get_path($path)))
			return true;

		return false;

	}


	public function merge ($source = "", $target = "") {

		return $this->copy($source, $target, ["merge"=>true]);

	}


	public function move ($source = "", $target = "") {

		if ($this->copy($source, $target))
			return $this->delete($source);

		return false;
	}


	public function name ($path = "", $include_dot_names = false) {

		if (is_empty($path))
			return $path;

		$path_parts = preg_split("/\?/",$path);

		$path = key_of(0,$path_parts);
		$path = preg_split("/\//",$path);

		if (!$include_dot_names && matches(".", current(array_slice($path, -1))))
			return current(array_slice($path, -2,1));

		if (is_empty(array_slice($path, -1)))
			return current(array_slice($path, -2,1));

		return current(array_slice($path, -1));

	}


	public function parent ($path = "") {

		if (any_empty($path, is_string($path)))
			return "";

		return either(
			strstr($this->part($path), $this->name($path), true),
			""
		);
	}


	public function part ($path = "") {

		if (is_empty($path))
			return $path;

		$path_parts = preg_split("/\?/",$path);

		$path = key_of(0,$path_parts);
		$path = preg_split("/\//",$path);

		if (matches(".", current(array_slice($path, -1))))
			$path = array_slice($path, 0, -1);

		$path = $this->core->array->trim($path);

		$path = ifttt($path, implode("/",$path)."/", "");

		if (matches($this->core->path("root"), "/".$path))
			$path = "/".$path;

		return $path;
	}


	public function read ($path = "", $options = []) {

		if (count($args = func_get_args()) == 1 && is_array($args[0]))
		  	return $this->read("", $args[0]);

		$options = $this->core->array->union(
			[
				"files" => true,
				"folders" => true,
				"name" => false,
				"extension" => false,
				"name_exclude" => false,
				"extension_exclude" => false,
				"stats" => false,
				"full_path" => false,
				"prepend" => ""
			],
			$options
		);


		// Backwards compatibility
		$options["extension"] = either(key_of("filetype",$options), $options["extension"]);
		$options["prepend"] = ifttt(key_of("path", $options), $this->path($path), "");

		if ($type = key_of("type", $options)) {

			deprecated("type is deprecated");

			switch ($type) {
				case "folder":
					$options["files"] = false;
					break;
				case "file":
					$options["folders"] = false;
			}

		}

		if (is_indexed($paths = $path)) {

			$return = [];

			foreach ($paths as $path)
				$return[] = $this->read($path, $options);

			return $return;

		}

		if (not_empty(either($options["extension"], $options["extension_exclude"])))
			$options["folders"] = false;

		if (!$this->exists($path))
			return [];

		$list = scandir($this->get_path($path));

		foreach ($list as $index => $entry) {

			if (match_any($entry, [".",".."])) {
				unset($list[$index]);
				continue;
			}

			if (!$options["folders"] && is_dir($this->get_path($this->path($path).$entry))) {
				unset($list[$index]);
				continue;
			}

			if (!$options["files"] && is_file($this->get_path($this->path($path).$entry))) {
				unset($list[$index]);
				continue;
			}

			if (not_empty($options["extension"]) && !fetch($options["extension"], $this->fs->file->extension($entry))) {
				unset($list[$index]);
				continue;
			}

			if (not_empty($options["extension_exclude"]) && fetch($options["extension_exclude"], $this->fs->file->extension($entry))) {
				unset($list[$index]);
				continue;
			}

			if (
				not_empty($options["name"]) &&
				ifttt(
					is_file($this->get_path($this->path($path).$entry)),
					!fetch($options["name"], $this->fs->file->basename($entry)),
					!fetch($options["name"], $this->name($entry))
				)
			) {
				unset($list[$index]);
				continue;
			}

			if (not_empty($options["name_exclude"]) && fetch($options["name_exclude"], $this->fs->file->basename($entry))) {
				unset($list[$index]);
				continue;
			}

			if ($options["full_path"])
				$entry_edit = $this->path($this->get_path($this->dir()),$path).$entry;

			else
				$entry_edit = $options["prepend"].$entry;

			if ($options["stats"] && is_file($file = $this->get_path($this->path($path).$entry)))
				$entry_edit = [
					"name" => $entry_edit,
					"mtime" => key_of("mtime", $stat = stat($file)),
					"ctime" => key_of("ctime", $stat),
					"size" => key_of("size", $stat),
				];

			$list[$index] = $entry_edit;

		}

		return $list;

	}


	public function zip ($folder = "", $name = false, $options = []) {

		if (is_empty($folder))
			return false;

		if (!is_dir($this->get_path($folder)))
			return false;

		if (is_empty($name))
			$name = $this->name($folder);

		$zip = $this->core->zip();
		$zip->filesystem->dir($this->get_path());

		return $zip->folder($folder, $name, $options);

	}

}


