<?php

class KB_html extends KB_class {

	private $core;

	function __construct($core) {
		$this->core = $core;
	}


	public function __call($tag, $arguments) {

		$arguments = $this->core->array->merge(array($tag), $arguments);
		$html = call_user_func_array(array($this, "tag"), $arguments);

		return $this->render($html);

	}


	public function parse($string = "") {

		$object = array();
		$string = preg_replace("/[\n\r]/","",$string);
		$case = true;
		$search_html_search = '/<([^\s>]+)[\s]*([^>]*)>(.*)/i';
		$search_html_complete = '/<([^\s>]+)[\s]*([^>]*)>(.*)<\/\\1>/i';
		$search_html_selfclose = '/<([^\s>]+)[\s]*([^>]*)>/i';
		$search_text = '/([^<]*)(.*)/';

		if (is_empty($string))
			return $string;

		while ($case) {

			//Add inner text to array
			preg_match($search_text,$string, $text);
			$string = preg_replace($search_text, "\\2", $text[2], 1);
			$text = $text[1];
			if(!empty($text) && preg_match("/\S/",$text)) $object[] = $text;

			//Add nested elements
			if(!empty($string)){
				preg_match($search_html_search,$string, $array);
				$count = count($object);
				if(isset($array[1])){
					//Find closing tag, differentiate with self closing elements
					if(!in_array(strtolower($array[1]),$this->config("empty_elements"))){
						$search_pos = true;
						$pos1 = 0;
						$pos2 = 0;
						while($search_pos){
							$pos2 = strpos($array[0],'</'.$array[1],$pos2);
							$pos1 = strpos($array[0],'<'.$array[1],$pos1);
							$start_string = isset($start_string) ? $start_string : $pos1;
							if(strpos($array[0],'<'.$array[1],$pos1+1) && strpos($array[0],'<'.$array[1],$pos1+1) < $pos2){
								$pos2++;
								$pos1++;
							}else{
								$end_string = strpos($array[0],'>',$pos2)+1;
								$search_pos = false;
							}
						}
						$string_new = substr($string,$start_string,$end_string-$start_string);
						$string = substr($string,$end_string);
						preg_match($search_html_complete,$string_new, $array);
					}else{
						preg_match($search_html_selfclose,$string,$array);
						$string = substr($string,strlen($array[0]));
					}
					if(isset($array[2]) && not_empty($array[2])){
						$properties = preg_replace("/([\s]+)/"," ",$array[2]);
						$properties = substr($properties,-1) == "/" ? substr($properties,0,-1) : $properties;
						$properties = preg_replace("/(\")\s([\w_-]+=)/","\\1|\\2",$properties);
						$properties = explode("|",$properties);
						foreach($properties as $key => $value){
							$value = explode("=",$value);
							if(count($value) > 2){
								$temp = array_shift($value);
								$value[1] = implode("=",$value);
								$value[0] = $temp;
								unset($temp);
							}
							if(isset($value[1])){
								$object[$count][$array[1]][$value[0]] = preg_replace("#([\"]?)([^\"]*)([\"]?)#","\\2",$value[1]);
							}
							else if(!empty($value[0])){
								$object[$count][$array[1]][$value[0]] = "";
							}
						}
					}
					if(isset($array[3]) && not_empty($array[3]) && preg_match($search_html_search, $array[3])){
						foreach($this->parse($array[3]) as $entry){
							$object[$count][$array[1]][] = $entry;
						}
					}
					else if(isset($array[3]) && not_empty($array[3])){

						$object[$count][$array[1]][] = stripslashes($array[3]);

					}
					else if(!empty($array[1]) && !isset($object[$count])){

						$object[$count][$array[1]] = array();

					}

				}
			}
			if(empty($string)){
				$case = false;
			}
		}
		return $object;
	}


	public function render($string = "", $options = [], $depth = 0) {

		if (is_string($string))
			return $string;

		if ($options && is_string($options)) {
			$options = string_to_array($options);
		}

		$options["type"] = either(key_of("type",$options), $this->config("type"));
		$markup = "";

		foreach ($string as $key => $value) {

			if (!is_array($value)) {
				$markup .= $value;
			}

			else if (!is_numeric($key)) {

				$properties = array();

				foreach($value as $key2 => $value2){
					if (!is_numeric($key2) && not_empty($key2)) {

						if (is_array($value2))
							$value2 = implode(" ", $value2);

						if ($value2 !== false && $value2 !== NULL)
							$properties[] = ($value2 === true) ? $key2 : $key2."=\"".$value2."\"";

						unset($value[$key2]);
					}
				}

				$content = $this->render($value,$options,$depth+1);
				$closing_tag = ifttt($this->core->array->match(key_of("type", $options), array("xhtml", "xml")) && $this->core->array->match($key,$this->config("empty_elements")), "/", "");

				$markup .=	"<".$this->name_space($key);

				if (not_empty($properties))
					$markup .= " ".implode(" ",$properties);

				$markup .= "$closing_tag>".$content;

				if (!in_array($key,$this->config("empty_elements")))
					$markup .= "</".$this->name_space($key).">";

				$markup .= key_of("spacing",$options);

			}

			else{
				$markup .= $this->render($value,$options,$depth);
			}
			$markup .= "\n";
		}
		return trim($markup);
	}


	public function property($string = false, $property = false, $value = false) {

		if (is_empty($property))
			return $string;

		if (is_array($property) && !is_indexed($property) && $value === false) {

			foreach ($property as $key => $value) {

				$string = $this->property($string, $key, $value);
			}
			return $string;
		}

		if (not_empty($property) && matches("=", $property) && is_empty($value)) {
			preg_match("/([\w]+)=\"([\w\s:]+)\"/",$property, $matches);
			$property = key_of(1,$matches);
			$value = key_of(2,$matches);
			return $this->property($string, $property, $value);

		}

		// if (is_string($string) && is_empty($string))
			// return $string;

		// Dont set or remove property
		if ($value === false && is_string($string)){

			// Remove property
			if (preg_match(preg_mod($property."=([\"\'])"),$string)) {

				preg_match(preg_mod($property."=(?:[\"\'])([^\"\'])"),$string,$return);
				return $return[1];
			}

			// Dont set property
			else {
				return $string;
			}
		}

		// Add property
		else {

			if (is_bool($string) || is_null($string))
				$string = "";

			// For rendered HTML
			if (is_string($string)) {

				if (preg_match(preg_mod($property."=([\"\'])"),$string)) {

					preg_match(preg_mod($property."=(?:[\"\'])([^\"\']+)"),$string,$return);

					if(!preg_match(preg_mod($property),$return[1]))
						$replace = $return[1]." ".$value;
					else
						$replace = $value;

					$replace = preg_replace(preg_mod($return[1]),$replace,$return[0]);
					$string = preg_replace(preg_mod($return[0]),$replace,$string);

					return $string;
				}

				else if (preg_match("/^</",$string)){

					return preg_replace("/^([^>]+)>/","$1"." ".$property.'="'.$value.'">',$string);
				}

				else {

					if (not_empty($string))
						$string .= " ";

					return $string.$property.'="'.$value.'"';
				}
			}

			// For HTML array
			else {

				foreach ($string as $key => $array) {

					// Found an object
					if (preg_match("/[a-zA-Z]/",$key) && is_array($array)) {

						// Dont set twice
						if(!fetch($value, key_of($property, $string[$key])))
							// Either set or add
							$string[$key][$property] = either(
								ifttt(is_empty(key_of($property,$string[$key])), $value),
								key_of($property,$string[$key])." ".$value
								);

						// Done and return
						return $string;
					}

					// Object not found, recursive restart
					else if(is_array($array)) {
						$string[$key] = $this->property($array, $property, $value);
					}

					// Fail
					else {
						continue;
					}

				}
				return $string;
			}
		}

	}


	public function add_class($string = false, $class = false){

		if (is_empty($class))
			return $string;

		return $this->property($string, "class", $class);

	}


	public function add_markup($string = false, $markup = false){

		if (any_empty($string, $markup))
			return $string;


		if (is_indexed($array = $markup)) {
			foreach ($array as $markup)
				$string = $this->add_markup($string, $markup);

			return $string;
		}

		if (is_string($markup)) {

			$markup = $this->parse("<br $markup>");
			$markup = key_fetch("br",$markup);
		}

		if (is_empty($markup))
			return $string;

		foreach ($markup as $property => $value)
			$string = $this->property($string, $property, $value);


		return $string;

	}


	public function tag($tag = "", $content = "", $properties = array()){

		if (is_empty($tag) || !is_string($tag))
			return array();

		if (fetch($tag,$this->config("empty_elements"))) {
			$properties = $content;
			$content = false;
		}

		if (
				is_string($content) &&
				matches("<",$content) &&
				substr_count("<",$content) != substr_count(">",$content)
			)
			$content = htmlspecialchars($content);


		if (!is_array($properties))
			$properties = array();

		return array(
			$tag => array(
				$content
			)+$properties
		);

	}


	public function table($table = [], $head = false, $properties = array()) {

		if (is_empty(key_of("table", $properties)))
			$properties = array(
				"table" => $properties
			);

		if (are_empty(
			!any_empty($table, is_array($table)),
			$head
		))
			return array();

		$count = 1;

		if (not_empty($head) && is_array($head))
			$count = count($head);

		foreach($table as $row)
			if (is_array($row) && ($row_count = count($row)) > $count)
				$count = $row_count;

		if(not_empty($head) && is_array($head)) {

			$row_count = count($head);

			$head_column = array();

			foreach ($head as $column) {
				$head_column[] = $this->tag("th",$column, array("colspan"=>ifttt($count - $row_count > 0, $count - $row_count + 1)));
				$row_count = $count;
			}

			$head = $this->tag("thead", $this->tag("tr",$head_column));
		}

		$body_row = array();

		foreach($table as $row){

			if (is_array($row)) {

				$row_count = count($row);

				$body_column = array();

				foreach ($row as $column) {

					$cell_properties = [];

					if (is_indexed($column)) {
						$cell_properties = key_of(1,$column);
						$column = key_of(0,$column);
					}

					$cell_properties = $this->core->array->merge($cell_properties,array("colspan"=>ifttt($count - $row_count > 0, $count - $row_count + 1)));

					$body_column[] = $this->tag("td",$column, $cell_properties);
					$row_count = $count;
				}

				$body_row[] = $this->tag("tr",$body_column);
			}

			else {
				$body_row[] = $this->tag("tr",$this->tag("td",$row));
			}
		}

		$body = $this->tag("tbody",$body_row);

		$return = $this->tag("table", array($head, $body), key_of("table", $properties));
		return $return;

	}


	public function img($src = "", $properties = array()) {

		$properties = $this->core->array->merge(
			$properties,
			array(
				"src" => $src
			)
		);

		return $this->render($this->tag("img", $properties));

	}


	private function name_space($tag = "") {

		if (matches(":", $tag))
			return $tag;

		$namespace = $this->config("namespace");

		return ifttt(
			$namespace,
			$namespace.":".$tag,
			$tag
		);

	}

}
