<?php

class KB_mail extends KB_class {

	private
		$core,
		$fs;

	function __construct($core) {

		$this->core = $core;

		$this->from(
			$this->core->config("admin_email"),
			$_SERVER["SERVER_NAME"]
		);

		$this->fs = $this->core->filesystem();
		$this->fs->base_dir($this->core->path("files"));
	}

	// Public

	public function attachment($file = "") {

		if (is_empty($file))
			return;

		if (is_indexed($files = $file, true)) {
			$return = array();
			foreach ($files as $file) {
				$return[] = $this->part("file", $file);
			}
			return $return;
		}

		return $this->part("file", $file);

	}


	public function bcc($set = NULL, $replace = false) {

		$to = $this->private_config("bcc");

		if (is_null($set)) {

			if (is_array($to))
				$to = implode(", ", $to);

			return $to;
		}

		if (is_email($set)) {

			if ($replace || is_empty($to))
				$this->private_config("bcc",$set);

			else
				$this->private_config("bcc", $this->core->array->merge((array) $to, (array) $set));

		}

		// Allow reset

		else if (is_empty($set) && $replace) {

			$this->private_config("bcc",$set);

		}

		return $this->bcc();

	}


	public function body($set = NULL, $replace = false) {

		if (is_null($set))
			return;

		if (are_empty(is_string($set), is_int($set)))
			return;

		$this->text(trim(strip_tags($set)), $replace);
		$this->html($set, $replace);
	}


	public function cc($set = NULL, $replace = false) {

		$to = $this->private_config("cc");

		if (is_null($set)) {

			if (is_array($to))
				$to = implode(", ", $to);

			return $to;
		}

		if (is_email($set)) {

			if ($replace || is_empty($to))
				$this->private_config("cc",$set);

			else
				$this->private_config("cc", $this->core->array->merge((array) $to, (array) $set));

		}

		// Allow reset

		else if (is_empty($set) && $replace) {

			$this->private_config("cc",$set);

		}

		return $this->cc();

	}


	public function compose() {

		$this->to("", true);
		$this->cc("", true);
		$this->bcc("", true);
		$this->from("", true);
		$this->body("", true);
		$this->reply("");
		$this->part("","",true);
		$this->header("","",true);

	}


	public function from($sender = NULL, $name = false) {

		$header = $this->private_config("header");

		if (is_null($sender))
			return key_of("From", $header);

		if (!is_email($sender))
			return false;

		$this->reply($sender);

		if (not_empty($name))
			$set = html_entity_decode(str_replace('\"', '"', $name), ENT_QUOTES, 'UTF-8').' <'.html_entity_decode(str_replace('\"', '"', $sender),ENT_QUOTES,'UTF-8').'>';

		else
			$set = html_entity_decode(str_replace('\"', '"', $sender),ENT_QUOTES,'UTF-8');

		$this->header("From", $set);

		return $set;
	}


	public function header($set = NULL, $value = NULL, $replace = false) {

		$header = $this->private_config("header");

		if (is_null($set))
			return $header;

		if ($replace)
			$header = array();

		if (is_empty($value))
			$header = remove_keys_of($set, $header);

		else
			$header[$set] = $value;

		$this->private_config("header", $header);
		return $header;
	}


	public function html($set = NULL, $replace = false) {

		$html = $this->private_config("html");

		if(is_null($set))
			return $html;

		$set = html_entity_decode(str_replace('\"', '"', $set),ENT_QUOTES,'UTF-8');
		$html = ifttt($replace, $set, $html.br.$set);
		$this->private_config("html", $html);

		return $html;

	}


	public function option($set = NULL, $value = false) {

		$options = $this->private_config("options");

		if (is_null($set))
			return $options;

		if (is_empty($value))
			$options = remove_keys_of($set, $options);

		else
			$options[$set] = $value;

		$this->private_config("options", $options);
		return $options;
	}


	public function part($type = NULL, $value = false, $replace = false) {

		$parts = $this->private_config("parts");

		if (is_null($type))
			return $parts;

		if ($replace)
			$parts = array();

		if (not_empty($type, $value))
			$parts[] = array(
				"type" => $type,
				"value" => $value
			);

		$this->private_config("parts", $parts);
		return $parts;

	}


	public function reply($set = NULL) {

		$reply = $this->private_config("reply");

		if (is_null($set))
			return $reply;

		if (is_email($set))
			$this->private_config("reply",$set);

		return $this->reply();

	}


	public function send($to = "", $subject = "", $message = "", $headers = array()){

		if (not_empty($to))
			$this->to($to, true);

		if (not_empty($subject))
			$this->subject($subject);

		if (not_empty($message))
			$this->body($message, true);

		if (is_empty($this->to()))
			$this->to($this->core->user->data("name"));

		if (is_empty($this->to()))
			return false;

		if(not_empty($this->part()))
			return $this->send_mime_mixed();

		if (not_empty($this->text()) && not_empty($this->html()))
			return $this->send_mime_alternative();

		if (not_empty($this->text()))
			return $this->send_text_mail();

		if (not_empty($this->html()))
			return $this->send_html_mail();

		return false;
	}


	public function subject($set = NULL){

		$subject = $this->private_config("subject");

		if(is_null($set))
			return $subject;

		$set = html_entity_decode($set,ENT_QUOTES);
		$subject = $set;
		$this->private_config("subject", $subject);

		return $subject;
	}


	public function text($set = NULL, $replace = false){

		$text = $this->private_config("text");

		if(is_null($set))
			return $text;

		$set = html_entity_decode(str_replace('\"', '"', $set),ENT_QUOTES,'UTF-8');
		$text = ifttt($replace, $set, $text.br.br.$set);
		$this->private_config("text", $text);

		return $text;
	}


	public function to($set = NULL, $name = NULL, $replace = NULL) {

		if (is_bool($name) && is_null($replace)) {
			$replace = $name;
			$name = "";
		}

		if (is_string($set) && not_empty($set) && preg_match("/^([^<]+)<([^>]+)>$/", $set, $matches) && is_email($matches[2])) {
			$set = trim($matches[2]);
			$name = trim($matches[1]);
		}

		$to = $this->private_config("to");

		if (is_null($set))
			return $to;


		if (is_indexed($sets = $set)) {

			foreach($sets as $set)
				$this->to($set, $replace);

			return $this->to();
		}

		if (is_email($set)) {

			if (not_empty($name))
				$set = html_entity_decode(str_replace('\"', '"', $name), ENT_QUOTES, 'UTF-8').' <'.html_entity_decode(str_replace('\"', '"', $set),ENT_QUOTES,'UTF-8').'>';


			if ($replace === true || is_empty($to))
				$this->private_config("to",$set);

			else
				$this->private_config("to", $this->core->array->merge((array) $to, (array) $set));

		}

		// Allow reset

		else if (is_empty($set) && $replace) {

			$this->private_config("to",$set);

		}

		return $this->to();

	}



	// Alias

	public function reset() {
		return $this->compose();
	}



	// Private

	private function get_to() {

		$to = $this->private_config("to");

		if (is_array($to))
			$to = implode(", ", $to);

		return $to;

	}

	private function private_header($set = NULL, $value = NULL) {

		$header = $this->private_config("private_header");

		if (is_empty($set))
			return $header;

		if (is_empty($value))
			$header = remove_keys_of($set, $header);

		else
			$header[$set] = $value;

		$this->private_config("private_header", $header);
		return $header;
	}

	private function send_text_mail(){

		$this->private_header("X-Mailer", "PHP/" . phpversion());
		$this->private_header("X-Sender-IP", $_SERVER['REMOTE_ADDR']);
		$this->private_header("Content-type", "text/plain; charset=utf-8");
		$this->private_header('Cc', $this->cc());
		$this->private_header('Bcc', $this->bcc());

		return mail(
			$this->get_to(),
			$this->get_subject(),
			$this->text(),
			$this->get_headers(),
			$this->get_options()
		);

	}

	private function send_html_mail() {

		$this->private_header("X-Mailer", "PHP/" . phpversion());
		$this->private_header("X-Sender-IP", $_SERVER['REMOTE_ADDR']);
		$this->private_header("MIME-Version", "1.0");
		$this->private_header("Content-type", "text/html; charset=utf-8");
		$this->private_header('Cc', $this->cc());
		$this->private_header('Bcc', $this->bcc());

		return mail(
			$this->get_to(),
			$this->get_subject(),
			$this->html(),
			$this->get_headers(),
			$this->get_options()
		);
	}

	private function send_mime_alternative() {

		$boundary = "=_NextPart-".time().rand(1,10);

		$this->private_header("X-Mailer", "PHP/" . phpversion());
		$this->private_header("X-Sender-IP", $_SERVER['REMOTE_ADDR']);
		$this->private_header('MIME-Version', '1.0');
		$this->private_header('Content-Type', 'multipart/alternative; boundary="'.$boundary.'"');
		$this->private_header('Cc', $this->cc());
		$this->private_header('Bcc', $this->bcc());

		$text = '--'.$boundary.PHP_EOL.
			'Content-Type: text/plain; charset=UTF-8'.PHP_EOL.
			'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
			str_replace(chr(13).chr(10), chr(10),quoted_printable_encode($this->text())).PHP_EOL.
			'--'.$boundary.PHP_EOL.
			'Content-Type: text/html; charset=UTF-8'.PHP_EOL.
			'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
			str_replace(chr(13).chr(10), chr(10),quoted_printable_encode($this->html())).PHP_EOL.
			'--'.$boundary.'--';

		return mail(
			$this->get_to(),
			$this->get_subject(),
			$text,
			$this->get_headers(),
			$this->get_options()
		);
	}

	private function send_mime_mixed() {

		$text = array();
		$boundary = "=_NextPart-".time().rand(1,10);

		$this->private_header("X-Mailer", "PHP/" . phpversion());
		$this->private_header("X-Sender-IP", $_SERVER['REMOTE_ADDR']);
		$this->private_header('MIME-Version', '1.0');
		$this->private_header('Content-Type', 'multipart/mixed; boundary="'.$boundary.'"');
		$this->private_header('Cc', $this->cc());
		$this->private_header('Bcc', $this->bcc());

		if(not_empty($this->html()) && not_empty($this->text())) {

			$boundary2 = "=_TextPart-".time().rand(1,10);
			$text[] = '--'.$boundary.PHP_EOL.
				'Content-Type: multipart/alternative;  boundary="'.$boundary2.'"'.PHP_EOL.
				'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
				'--'.$boundary2.PHP_EOL.
				'Content-Type: text/plain; charset="utf-8"'.PHP_EOL.
				'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
				imap_8bit($this->text()).PHP_EOL.
				'--'.$boundary2.PHP_EOL.
				'Content-Type: text/html; charset=UTF-8'.PHP_EOL.
				'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
				imap_8bit($this->html()).PHP_EOL.
				'--'.$boundary2.'--';
		}

		else if(not_empty($this->text())) {
			$text[] =  '--'.$boundary.PHP_EOL.
				'Content-Type: text/plain; charset="utf-8"'.PHP_EOL.
				'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
				imap_8bit($this->text()).PHP_EOL;
		}

		else if(not_empty($this->html())) {
			$text[] =  '--'.$boundary.PHP_EOL.
				'Content-Type: text/html; charset=UTF-8'.PHP_EOL.
				'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
				imap_8bit($this->text()).PHP_EOL;
		}


		foreach ($this->part() as $part) {

			if (any_empty(key_of("type", $part), key_of("value", $part)))
				continue;

			if ($part['type'] == 'text') {
				$text[] =  '--'.$boundary.PHP_EOL.
					'Content-Type: text/plain; charset="utf-8"'.PHP_EOL.
					'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
					imap_8bit($part['value']);
			}

			else if($part['type'] == 'html') {
				$text[] = '--'.$boundary.PHP_EOL.
					'Content-Type: text/html; charset=UTF-8'.PHP_EOL.
					'Content-Transfer-Encoding: quoted-printable'.PHP_EOL.PHP_EOL.
					imap_8bit($part['value']);
			}

			else if ($part['type'] == 'file') {

				if (!$this->fs->exist_file($part['value']))
					continue;

				$content_type =  $this->core->file->type($part['value']);

				$text[] = '--'.$boundary.PHP_EOL.
					'Content-Type: '.$content_type.' name="'.$this->core->file->name($part['value']).'"'.PHP_EOL.
					'Content-Transfer-Encoding: base64'.PHP_EOL.
					'Content-Disposition: attachment; filename="'.$this->core->file->name($part['value']).'"'.PHP_EOL.PHP_EOL.
					chunk_split(base64_encode($this->fs->read_file($part['value'])));

			}
		}

		$text = implode(PHP_EOL,$text);

		return mail(
			$this->get_to(),
			$this->get_subject(),
			$text,
			$this->get_headers(),
			$this->get_options()
		);
	}

	private function get_headers() {

		$header = $this->core->array->merge(
			$this->private_config("header"),
			$this->private_config("private_header")
		);

		$this->private_config("private_header", array());

		if (is_empty($header))
			return "";

		$head = array();

		foreach($header as $key => $value) {

			if(is_numeric($key))
				continue;

			$head[] = $key.": ".$value.PHP_EOL;
		}

		return implode($head);

	}

	private function get_subject() {

		if (!defined("HEAD_CRLF"))
			define("HEAD_CRLF","\r\n");

		$encoding = "UTF-8";

		$string = str_replace(" ", "_", trim($this->private_config("subject"))) ;

		// We need to delete "=\r\n" produced by imap_8bit() and replace '?'
		if (loaded("imap"))
    	$string = str_replace("?", "=3F", str_replace("=\r\n", "", imap_8bit($string))) ;

    // Now we split by \r\n - i'm not sure about how many chars (header name counts or not?)
    //$string = chunk_split($string, 73);
    $string = chunk_split($string, 1000);
    // We also have to remove last unneeded \r\n :
    $string = substr($string, 0, strlen($string)-2);
    // replace newlines with encoding text "=?UTF ..."
    $string = str_replace("\r\n", "?=".HEAD_CRLF." =?".$encoding."?Q?", $string) ;

    return '=?'.$encoding.'?Q?'.$string.'?=';
	}

	private function get_options() {

		$options = $this->private_config("options");

		if (not_empty($reply = $this->reply()))
			$options[] = "-f ".$reply;

		if (is_empty($options))
			return "";

		return implode(" ",$options);
	}



	// Deprecated

	public function add_mixed_part($type = NULL, $value = false, $replace = false) {
		deprecated("Use ->part() instead of ->add_mixed_part()");
		return $this->part($type, $value, $replace);
	}

	public function add_sendmail_options($set = NULL, $value = false) {
		deprecated("Use ->option() instead of ->add_sendmail_options()");
		if(is_array($options = $set))
			foreach($options as $set => $value)
				$this->option($set, $value);
		else
			$this->option($set, $value);
	}

	public function set_env_sender($set = NULL) {
		deprecated("Use ->reply() instead of ->set_env_sender()");
		return $this->reply($set);
	}

	public function set_sender($sender = NULL, $name = false) {
		deprecated("Use ->from() instead of ->set_sender()");
		return $this->from($sender, $name);
	}

	public function add_recipient_bcc($set = NULL){
		deprecated("Use ->bcc() instead of ->add_recipient_bcc()");
		return $this->bcc($set);
	}

	public function add_recipient_cc($set = NULL) {
		deprecated("Use ->cc() instead of ->add_recipient_cc()");
		return $this->cc($set);
	}

	public function set_recipient($set = NULL) {
		deprecated("Use ->to(,true) instead of ->set_recipient()");
		return $this->to($set, true);
	}

	public function add_recipient($set = NULL) {
		deprecated("Use ->to() instead of ->add_recipient()");
		return $this->to($set);
	}

	public function add_html($set = "") {
		deprecated("Use ->html() instead of ->add_html()");
		return $this->html($set);
	}

	public function add_text($set = "") {
		deprecated("Use ->text() instead of ->add_text()");
		return $this->text($set);
	}

	public function add_body($body = false){
		deprecated("Use ->body() instead of ->add_body()");
		$this->body($body);
	}

	public function send_mail() {
		deprecated("Use ->send() instead of send_mail()");
		return $this->send();
	}

	public function add_body_txt($set = "") {
		deprecated("Use ->text() instead of ->add_body_txt()");
		return $this->text($set);
	}

	public function set_body_txt($set = NULL) {
		deprecated("Use ->text(,true) instead of ->set_body_txt()");
		return $this->text($set, true);
	}

	public function add_body_html($set = "") {
		deprecated("Use ->html() instead of ->add_body_html()");
		return $this->html($set);
	}

	public function set_body_html($set = NULL) {
		deprecated("Use ->html(,true) instead of ->set_body_html()");
		return $this->html($set, true);
	}

	public function set_subject($set = NULL) {
		deprecated("Use ->subject() instead of ->set_subject()");
		return $this->subject($set);
	}

	public function set_header($key = NULL, $value = NULL) {
		deprecated("Use ->header() instead of ->set_header()");
		return $this->header($key, $value);
	}

	public function set_headers($set) {
		deprecated("set_headers is deprecated");

		if(is_array($set))
			foreach ($set as $key => $value)
				$this->header($set, $value);
		return;
	}


}
