<?php

class KB_pagination extends KB_class {

	public $identifier = "";
	public $pagination = false;

	private $page = 0;
	private $array_identifier = "";
	private $count = 0;

	private $core;

	function __construct($core) {

		$this->core = $core;

	}
	// private $config = array("pagination" => 25);

	public function render($options=[]) {

		return $this->get_pagination($this->count, $options);

	}

	public function set($data = [], $pagination = false) {

		if (is_empty($data))
			return;

		// Object based pagination
		if (is_object($data) && $data instanceof  KB_data) {

			if (
				is_empty($data->search()) &&
				(!$data->properties() || is_empty($data->properties->filter()))
			)
				$this->count = $data->count();

			else
				$this->count = count($data->read());

			$this->pagination = either($pagination, false);
			$this->array_identifier = $this->identifier.md5(serialize(array($data->db_table(),$data->filter())));

			$page = $this->array_identifier.'_page';

			if (not_empty($this->core->get($page)) && ($this->core->get($page) * $this->pagination > $this->count)) {
				$this->core->set($page,0);
			}

			$data->limit($this->get_start(),$this->pagination);

			return $data;
		}

		// Data based pagination
		else {
			$this->count = count($data);

			if ($pagination == 0)
				$pagination = false;

			$this->pagination = $pagination;

			$this->array_identifier = $this->identifier.md5(serialize(array_keys($data)));

			if (!$this->pagination)
				return $data;

			$page = $this->array_identifier.'_page';

			if (not_empty($this->core->get($page)) && ($this->core->get($page) * $this->pagination > $this->count)) {
				$this->core->set($page,0);
			}

			$data = array_slice($data, $this->get_start(), $this->pagination);
			return $data;
		}
	}

	private function get_pagination($resCount, $options = []) {

		$options = array_multimerge(
			array(
				"info" => false,
				"page_links" => array(
					"type" => "a",
					"class" => ""
				),
				"jump_links" => false,
				"prev"=>true,
				"next"=>true,
				"first"=>true,
				"last"=>true,
			),$options
		);

		if ($this->core->runs("backend")) {
			$options["page_links"]["class"] = "small button";
			$options["active"] = array(
				"class" => "small button active"
			);
			$options["prev"] = array(
				"class" => "small button"
			);
			$options["next"] = array(
				"class" => "small button"
			);
			$options["first"] = array(
				"class" => "small button"
			);
			$options["last"] = array(
				"class" => "small button"
			);
			$options["info"] = array(
				"class" => "info pagination"
			);
			$options["jump_links"] = true;
		}

		$resCount = intval($resCount);

		if (any_empty($resCount, $this->pagination))
			$pages = 0;

		else
			$pages = (int)ceil($resCount/$this->pagination);

		$page_links = [];
		$page_links_before_current = [];
		$page_link_current = [];
		$page_links_after_current = [];

		if (not_empty($icon_next = key_fetch("icon",key_fetch("next",$options))))
			$icon_next = $this->core->html->img($icon_next);

		if (not_empty($icon_first = key_fetch("icon",key_fetch("first",$options))))
			$icon_first = $this->core->html->img($icon_first);

		if (not_empty($icon_prev = key_fetch("icon",key_fetch("prev",$options))))
			$icon_prev = $this->core->html->img($icon_prev);

		if (not_empty($icon_last = key_fetch("icon",key_fetch("last",$options))))
			$icon_last = $this->core->html->img($icon_last);

		if ($pages > 1) {

			$i = 0;
			$j = 0;
			$k = 1;

			if (key_fetch("jump_links",$options))
			while ($this->page - $i > 0) {

				$j = 1 - $j;
				$i = $j == 0 ? $k : 5 * $k;
				$k = $j == 0 ? $k : 10 * $k;

				if ($this->page-$i >= 0)
					$page_links[] = array(
						key_fetch("type",key_fetch("page_links",$options)) =>
						 	array(
								"href" => $this->gen_page_url($this->page-$i),
								"class" => key_fetch("class",key_fetch("page_links",$options)),
								"-".$i
							)
					);
			}

			if ($this->page > 0 && key_fetch("prev",$options))
				$page_links[] =	array(
					"a" => array(
						"href" => $this->gen_page_url($this->page - 1),
						"class" => key_fetch("class",key_fetch("prev",$options)),
						either(
							$icon_prev,
							$this->core->template->icon("prev")
						)
					)
				);

			if ($this->page > 1 && key_fetch("first",$options))
				$page_links[] =	array(
					"a" => array(
						"href" => $this->gen_page_url(0),
						"class" => key_fetch("class",key_fetch("first",$options)),
						either(
							$icon_first,
							$this->core->template->icon("first")
						)
					)
				);

			$page_links = array_reverse($page_links);

			for ($i = $this->page - 9; $i < $this->page; $i++) {

				if ($i >= 0)
					$page_links_before_current[] = array(
						key_fetch("type",key_fetch("page_links",$options)) =>
						 	array(
								"href" => $this->gen_page_url($i),
								"class" => key_fetch("class",key_fetch("page_links",$options)),
								($i+1)
							)
					);
			}

			$page_link_current[] = array(
				"a" => array(
					"class" => key_fetch("class",key_fetch("active",$options)),
					$this->page+1
				)
			);

			for ($i = $this->page+1; $i < $pages && $i <= $this->page + 9; $i++) {
				$page_links_after_current[] = array(
					key_fetch("type",key_fetch("page_links",$options)) =>
						array(
							"href" => $this->gen_page_url($i),
							"class" => key_fetch("class",key_fetch("page_links",$options)),
							($i+1)
						)
				);
			}

			// Limit pagination to 5 entries

			$before = count($page_links_before_current);
			$after = count($page_links_after_current);

			// Slice both sides
			if ($before >= 2 && $after >= 2) {
				$page_links_before_current = array_slice($page_links_before_current,-2);
				$page_links_after_current = array_slice($page_links_after_current,0,2);
			}
			// Slice after current
			else if ($before < 2 && $after >= 2) {
				$page_links_after_current = array_slice($page_links_after_current,0,4-$before);
			}
			// Slice before current
			else if ($before >= 2 && $after < 2) {
				$page_links_before_current = array_slice($page_links_before_current,-(4-$after));
			}


			$page_links[] = $page_links_before_current;
			$page_links[] = $page_link_current;
			$page_links[] = $page_links_after_current;


			$i = 0;
			$j = 0;
			$k = 1;

			if (key_fetch("jump_links",$options))
				while ($i + $this->page < $pages) {

					$j = 1 - $j;
					$i = $j == 0 ? $k : 5 * $k;
					$k = $j == 0 ? $k : 10 * $k;

					if ($i + $this->page < $pages)
						$page_links[] = array(
							key_fetch("type",key_fetch("page_links",$options)) =>
							 	array(
									"href" => $this->gen_page_url($i+$this->page),
									"class" => key_fetch("class",key_fetch("page_links",$options)),
									"+".$i
								)
						);
				}

			if ($this->page+1 < $pages && key_fetch("next",$options))
				$page_links[] =	array(
					"a" => array(
						"href" => $this->gen_page_url($this->page + 1),
						"class" => key_fetch("class",key_fetch("next",$options)),
						either(
							$icon_next,
							$this->core->template->icon("next")
						)
					)
				);

			if ($this->page+2 < $pages && key_fetch("last",$options))
				$page_links[] =	array(
					"a" => array(
						"href" => $this->gen_page_url($pages-1),
						"class" => key_fetch("class",key_fetch("last",$options)),
						either(
							$icon_last,
							$this->core->template->icon("last")
						)
					)
				);


			$page_links = array(
				"div" => array(
					"class" => "pagination links fill flex",
					$page_links,
				)
			);
		}

		$info = [];

		if (not_empty($page_links) && key_fetch("info",$options)) {

			$from = ($this->pagination * ($this->page)) +1;
			$to = ($this->pagination * ($this->page + 1));

			if ($to > $resCount)
				$to = $resCount;

			$info[] =	$this->core->html->tag("span", $this->core->language->translate("Page ".($this->page + 1)." of ".$pages)." ");
			$info[] =	$this->core->html->tag("span", $this->core->language->translate("(Entry ".$from." to ".$to." of total ".$resCount.")"));

			$info = [
				"div" => [
					"class" => either(key_fetch("class",key_fetch("info",$options)),"pagination_info"),
					$info,
				]
			];
		}

		return $this->core->html->render([$info, $page_links]);
	}

	private function gen_page_url($page) {
		$set_var = $this->array_identifier.'_page';
		return $this->core->generate_url(array("set_var"=>$set_var,"set_value"=>$page));
	}

	function get_start() {

		$page = $this->array_identifier.'_page';

		$page = either($this->core->get($page), 0);
		$this->page = $page;

		return $page * $this->pagination;
		return 0;
	}

}