<?php

class KB_user extends KB_data {

	private $user_data_array = false;
	private $group_data_array = false;

	private $core;

	function __construct($core) {

		parent::__construct($core);
    $this->core = $core;

		$this->db_table("kb_user", array(
      "name",
      "password",
      "active" => "true",
      "archived" => "false",
      "lastlogin" => "datetime DEFAULT NULL",
      "regdate" => "timestamp NULL DEFAULT CURRENT_TIMESTAMP COMMENT 'readonly'",
      "session",
    ));

    $this->order("name");
    $this->statistic(true);
    $this->properties(true);

		$this->group = new KB_user_group($core);

	}


	// Public

	public function backend () {

		if (!is_null($cache = $this->private_config("backend")))
			return $cache;

		if (!$this->data())
			return $this->private_config("backend", false);

		if (!$this->data("active"))
			return $this->private_config("backend", false);

		if ($this->data("archived"))
			return $this->private_config("backend", false);

		if (!$this->group->data("backend"))
			return $this->private_config("backend", false);

		$parentgroups = $this->group->read_parents($this->data("group_ID"));

		// Create new object to read all parents
		$usergroup = $this->core->data_group("kb_user_group");

		foreach ($parentgroups as $ID) {

			if ($ID == "NULL")
				break;

			$data = $usergroup->entry($ID);

			if (!are_equal(key_of("backend", $data), "1"))
				return $this->private_config("backend", false);
		}

		return $this->private_config("backend", true);
	}


	public function check ($user = "", $password = "") {

		if (any_empty($user, $password))
			return false;

		if (strlen($password) > 4096)
			return false;

		$ldap_bind = false;

		if($this->core->config("ldap_auth") == 1) {

			$ldap = $this->core->ldap_o();

			if (
				($user_ID = $ldap->test_user_data($user, $password)) &&
				(not_empty($this->read(["ID"=>$user_ID,"active"=>"1","archived"=>"0"])))
			)
				$ldap_bind = true;

		}

		if (!$ldap_bind)
			$entry = $this->entry([
				"name"=>$user,
				"password"=>$this->create_passwords($password),
				"active"=>"1",
				"archived"=>"0",
			]);

		if (!empty($entry) || $ldap_bind) {

			$user_ID = (!empty($entry)) ? $entry['ID'] : false;
			$this->core->session->cookie("kitstore",md5($password));
			$this->user_login($user_ID);
			return $user_ID;
		}

		return false;
  }


	public function create($array = array()){

		if (any_empty($array, key_of("name", $array), key_of("group_ID", $array), key_of("password", $array)))
			return false;

		if ($this->exists($array["name"],$array["group_ID"]))
			return false;


		$array["password"] = $this->create_password($array["password"]);
		$array["active"] = "1";

		return parent::create($array);

	}


	public function current () {

		if (!is_null($user = $this->private_config("current")))
			return $user;

		if (!$session_ID = $this->core->session->id())
				return $this->private_config("current", false);

		// Read user via session
		$user = $this->entry(["session" => $session_ID]);

		if (is_empty($user))
			return $this->private_config("current", false);

		return $this->private_config("current", $user);

	}


	public function create_password ($password) {

		$hash = md5($password);
		$cycles = $this->core->config("password_cycles");

		for ($i = $cycles; $i > 0; $i--)
			$hash = hash('sha256',$i.$hash);

		return $hash;
	}


	public function data ($key = false, $user = NULL) {

		if (not_empty($user))
			$user = $this->entry($user);

		else
			$user = $this->current();

		if (is_empty($user))
			return NULL;

		return ifttt($key, key_of($key, $user), $user);

	}


	public function exists ($name, $group = false) {

		if ($group)
			$user = $this->entry(
				array(
					"name"		=> $name,
					"group_ID"	=> $group
				)
			);

		else
			$user = $this->entry(
				array(
					"name"		=> $name,
				)
			);

		return not_empty($user);
	}


	public function ID () {
		return $this->data("ID");
	}


	public function is_dev () {

		if (!is_null($cache = $this->private_config("is_dev")))
			return $cache;

		if (not_empty($this->group->data()) && is_null($this->group->data('parent_ID')))
			return $this->private_config("is_dev", true);

		return $this->private_config("is_dev", false);
	}


	public function login ($ID = "") {

		if (any_empty($ID, is_numeric($ID)))
			return;

		$user = $this->entry($ID);

		if (are_equal(key_of("active", $user), 0))
			$user = [];

		if (are_equal(key_of("archived", $user), 1))
			$user = [];

		$this->private_config("current",$user);

		if (not_empty($this->group->data()) && is_null($this->group->data('parent_ID')))
			$this->private_config("is_dev", true);

		else
			$this->private_config("is_dev", false);

		return $user;

	}


	public function logged_in (){
		return not_empty($this->data());
	}


	public function sign_out() {

		if (!$this->logged_in())
			return true;

		$user = fetch($this->core->session->id(),$this->read());

		if (is_empty($user))
			return true;

		$user = current($user);

		$session = array_find($this->core->session->id(),$user);

		$update = array(
			"ID" => $user["ID"],
			key($session) => "",
		);

		$this->statistic(false);
		$this->update($update);

		return true;
	}


	public function update ($array = array(), $where=array()) {

		if(isset($array["password"]))
		$array["password"] = $this->create_password($array["password"]);

		return parent::update($array,$where);

	}



	// Private

	private function create_passwords($password) {

		$cycles = $this->core->config("password_cycles");

		// VER 1

		$hash = md5($password);
		$passwords[] = $hash;

		// VER 2

		for ($i = $cycles; $i > 0; $i--)
			$hash = hash('sha256',$password.$i.$hash);

		$passwords[] = $hash;

		// VER 3

		$passwords[] = $this->create_password($password);

		/* VER 3 upgrade from VER 1*/

		$hash = $password;

		for ($i = $cycles; $i > 0; $i--)
			$hash = hash('sha256',$i.$hash);

		$passwords[] = $hash;

		return $passwords;
	}

	private function user_login($ID){

		// Exlude API Calls
		if(!isset($_SERVER['HTTP_USER_AGENT'])){
			$this->user_data_array = $this->entry($ID);
			return true;
		}

		$this->update(
			array(
				"session" => "",
			),
			array(
				"session" => $this->core->session->id(),
			)
		);


		$this->update(
			array(
				"ID" => $ID,
				"session" => $this->core->session->id(),
				"lastlogin" => date("Y-m-d H:i:s"),
			)
		);
		$this->core->set_url_options(array("username"=>"","password"=>""));
		$this->core->set_var("username","");
		$this->core->set_var("password","");
    // $this->core->refresh = false;
	}



	// Deprecated

	public function all_user_groups() {

		deprecated();

		//Reads all user groups

		// $sql="SELECT ID,name FROM ".DB_PREFIX."kb_user_group";
		// $result = $this->db->sql_all($sql);

		$entries = $this->group->read_all();

		// $this->user_groups_indexed = $result;
		foreach($entries as $key => $value)
		{
			$this->user_groups[$value["ID"]] = $value["name"];
		}

	}
	public function get_user_data_per_ID($ID,$key){
		deprecated();
		return $this->data($key, $ID);
	}
	public function get_profile_col_num() {
		deprecated("get_profile_col_num is deprecated");
		if(isset($this->col_num) and !empty($this->col_num))
		{
			return $this->col_num;
		}
		else
		{
			$sql = "SHOW COLUMNS FROM ".DB_PREFIX."kb_user_profile";
			if($this->col_num = $this->db->num_rows($sql))
			{

				return $this->col_num;
			}
			return false;
		}
	}
	public function get_profile_cols() {
		deprecated();
		$sql = "SHOW COLUMNS FROM ".DB_PREFIX."kb_user_profile";
		if($cols = $this->db->sql_all($sql))
		{
			return $cols;
		}
		return false;

	}
	public function group_data($data = NULL){
		deprecated('Use $core->user->group->data()');
		return $this->group->data($data);
	}
	public function logmeout() {
		deprecated('Use $core->user->sign_out() instead of $core->user->logmeout()');
		return $this->sign_out();
	}
	public function set_backendlogin($ID, $login){

		deprecated();

		if(empty($login))
			$this->group->update(array("backend"=>"0","ID"=>$ID));

		else
			$this->group->update(array("backend"=>"1","ID"=>$ID));

		return true;
	}
	public function set_child_group($id) {

		deprecated();

	}
	public function set_profile_data($data = [], $id = NULL) {
		deprecated();
		foreach($data AS $a)
		{
			$cols[] = $a[0];
			$values[] = $a[1];
			$upd[] = "`".$a[0]."`=VALUES(`".$a[0]."`)";
		}

		$sql_user_data ="
			INSERT INTO ".DB_PREFIX."kb_user_profile (user_ID, `".implode("`,`", $cols)."`)
			VALUES ('".$id."', '".implode("', '", $values)."')
			ON DUPLICATE KEY UPDATE ".implode(", ", $upd);

		if($this->db->sql_query($sql_user_data, "set_profile_data"))
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	public function set_user_data($data=array(), $id='NULL') {
		deprecated();
		$sql_user_data ="
			INSERT INTO ".DB_PREFIX."kb_user (ID,group_ID,name,password,active)
			VALUES (".$id.", '".$data['group_ID']."','".$data['name']."','".$this->create_password($data['password'],$this->core->core_config["core"]["password_cycles"])."','0')
			ON DUPLICATE KEY UPDATE name=VALUES(name), password=VALUES(password)
			";
		$sql_recent_ID= "SELECT ID FROM ".DB_PREFIX."kb_user WHERE name = '".$data['name']."' LIMIT 1";

		if($this->db->sql_query($sql_user_data))
		{
			if($rec_id = $this->db->sql_first($sql_recent_ID)){ return $rec_id['ID']; }
		}
		else
		{
			return false;
		}
	}
	public function user_check($user = "", $password = "") {
		deprecated('Use $core->user->check() instead of $core->user->user_check()');
		return $this->check($user, $password);
	}
	public function user_data($key){
		deprecated("use ->data() instead of ->user_data()");
		return $this->data($key);

	}
	public function user_exists($name,$group=false) {
		deprecated();
		return $this->exists($name,$group);
	}
	public function user_ID(){
		deprecated('use $core->user->ID()');
		return $this->ID();
	}
	public function usergroup_exists($name){

		deprecated();
		return $this->group->exists($name);

	}

}


class KB_user_group extends KB_data_group {

	private $core;

	function __construct($core) {

		parent::__construct($core);
		$this->core = $core;

		$this->db_table("kb_user_group",array(
      "name",
      "backend"=>"tinyint(1) unsigned DEFAULT '0'",
			"parent_ID"=>"int(10)",
    ));

		$this->indexed(true);

	}


	public function current() {

		if (not_empty($group = $this->private_config("current")))
			return $group;

		if (is_empty($user = $this->core->user->current()))
			return array();

		// Read user via session
		$group = $this->entry_select(array("ID" => key_of("group_ID", $user)));

		if (is_empty($group))
			return array();

		$this->private_config("current", $group);

		return $group;

	}

	public function data($key = false) {

		if (not_empty($user = $this->current()))
			return ifttt($key, key_of($key, $user), $user);

		return NULL;

	}

	public function exists($name) {

		$group = $this->entry(
			array(
				"name"		=> $name,
			)
		);

		return not_empty($group);
	}

	public function ID() {
		return $this->data("ID");
	}
}