<?php

class KB_zip extends KB_class {

	public
		$filesystem;

	private
		$core,
		$zip;

	function __construct ($core) {

		$this->core = $core;
		$this->zip = false;
		$this->filesystem = $this->core->filesystem();
		$this->filesystem->dir($this->core->filesystem->base_dir());
	}



	// Public

	public function add_file ($file = "", $path = "", $content = NULL) {

		if (!$zip = $this->zip())
			return false;

		if (is_indexed($files = $file)) {
			foreach ($files as $file)
				$this->add_file($file, $path);

			return true;
		}

		if (any_empty($file, is_string($file)))
			return false;

		$file_name = $this->filesystem->file->name($file);

		$path = $this->filesystem->path($path);

		if (is_null($content))
			$zip->addFile($this->filesystem->get_path($file), $path.$file_name);

		else
			$zip->addFromString($path.$file_name, $content);

		if (method_exists($zip, "setCompressionName"))
			$zip->setCompressionName($path.$file_name, ZipArchive::CM_STORE);


		return true;

	}


	public function add_folder ($folder = "", $path = "", $options = []) {

		if (!$zip = $this->zip())
			return false;

		$options = $this->core->array->merge([
			"include_path" => false,
		], $options);

		if (is_indexed($folders = $folder)) {
			foreach ($folders as $folder)
				$this->add_folder($folder, $path);

			return true;
		}

		if (any_empty($folder, is_string($folder)))
			return false;

		$zip_path = $this->filesystem->path(
			$path,
			ifttt(
				key_of("include_path", $options),
				$this->filesystem->folder->part($folder),
				$this->filesystem->folder->name($folder)
			)
		);

		$zip->addEmptyDir($zip_path);

		foreach ($this->filesystem->folder->read($folder, ["files"=>false]) as $sub_folder)
			$this->add_folder($this->filesystem->path($folder, $sub_folder), $zip_path);

		foreach ($this->filesystem->folder->read($folder, ["folders"=>false]) as $file)
			$this->add_file($this->filesystem->path($folder).$file, $zip_path);

		return true;

	}


	public function close () {

		if (!$zip = $this->zip())
			return false;

		$zip->close();
	}


	public function dir ($set = NULL) {

		return $this->filesystem->dir($set);

	}


	public function extract ($file = "", $path = "") {

		if (any_empty($file, is_string($path)))
			return false;

		if (!$this->filesystem->file->exists($file))
			return false;

		$this->filesystem->folder->create($path);

		$zip = $this->zip();

		if (!$zip->open($this->filesystem->get_path($file)))
			return false;

		$zip->extractTo($this->filesystem->get_path($path));

		$zip->close();

		return true;
	}


	public function folder ($folder = "", $zip_name = "", $options = []) {

		if (!$zip = $this->zip())
			return false;

		if (any_empty($folder, $zip_name, is_string($folder), is_string($zip_name)))
			return false;

		$options = $this->core->array->merge([
			"path" => "",
			"include_path" => false,
		], $options);

		$zip_name = $this->create($zip_name);

		$this->add_folder($folder, key_of("path", $options), remove_keys_of("path", $options));

		$zip->close();

		return $zip_name;

	}


	public function create ($name = "") {

		if (any_empty($name, is_string($name)))
			return false;

		$zip = $this->zip();

		$name = $name.".zip";
		$name = $this->filesystem->get_path($name);

		$zip->open($name, ZipArchive::CREATE | ZipArchive::OVERWRITE);
		$zip->addEmptyDir(".");

		return $name;

	}



	// Private

	private function zip() {

		if ($this->zip && $this->zip instanceof ZipArchive)
			return $this->zip;

		$this->zip = new ZipArchive;

		return $this->zip;

	}



	// Deprecated

	public function addDir($path) {

		deprecated();

		print 'adding ' . $path . '<br>';
		$this->addEmptyDir($path);
		$nodes = glob($path . '/*');
		foreach ($nodes as $node) {
				print $node . '<br>';
				if (is_dir($node)) {
						$this->addDir($node);
				} else if (is_file($node))	{
						$this->addFile($node);
				}
		}
		return true;
	}
	public function addFolderToZip($dir, $path='', $zipdir = ''){

		deprecated();

		$rpath = '/'.trim($path, "\t\n\r/.").'/'.trim($dir, "\t\n\r/.");
			if (is_dir($rpath)) {
					if ($dh = opendir($rpath)) {
							$this->addEmptyDir($dir);

							while (($file = readdir($dh)) !== false) {

									//If it's a folder, run the function again!
									if (is_dir($rpath .'/'. $file)){
											if ( ($file !== ".") && ($file !== "..")){
													$this->addFolderToZip(trim($dir, "\t\n\r/.") .'/'. $file . "/", $path, $zipdir . $file );
											}
									}else{
											// Add the files
											//echo "<pre>adding: ".$rpath .'/'. $file." - ".trim($dir, "\t\n\r/.").'/'. $file."</pre>";
											$this->addFile($rpath .'/'. $file, trim($dir, "\t\n\r/.").'/'. $file);
									}
							}
					}
			}
		else
		{
			$this->addFile($rpath,trim($dir));
		}
	}
	public function error_code($err) {
		$errors = array(
			'10' => 'ZIPARCHIVE::ER_EXISTS',
			'21' => 'ZIPARCHIVE::ER_INCONS',
			'18' => 'ZIPARCHIVE::ER_INVAL',
			'14' => 'ZIPARCHIVE::ER_MEMORY',
			'9' => 'ZIPARCHIVE::ER_NOENT',
			'19' => 'ZIPARCHIVE::ER_NOZIP',
			'11' => 'ZIPARCHIVE::ER_OPEN',
			'5' => 'ZIPARCHIVE::ER_READ',
			'4' => 'ZIPARCHIVE::ER_SEEK'
			);
		return (isset($errors[$err])) ? $errors[$err] : false;
	}
	public function files($files = [], $options = [], $zip_o = false) {

		deprecated();

		if (is_empty($files))
			return false;

		if (is_string($files))
			$files = array($files);

		if (!is_array($files))
			return false;

		$options = $this->core->array->merge(
			array(
				"name" => "file",
				"folder" => false,
				"path" => DROOT.FILES_FOLDER,

			),
			$options
		);

		if ($zip_o && $zip_o instanceof ZipArchive) {
			$zip = $zip_o;
		}

		else {

			$zip = new ZipArchive;
			$dir = sys_get_temp_dir();

			if (is_empty($dir))
				return false;

			$dir .= "/";
			$zip_file = $dir.key_of("name", $options).".zip";
			$result = $zip->open($zip_file, ZipArchive::CREATE | ZipArchive::OVERWRITE);

			if (!$result)
				return false;
		}

		foreach ($files as $file) {
			$filename = $this->core->file->name($file);
			if (is_file($file = key_of("path", $options).$file))
				$zip->addFile($file, $filename);
		}

		if ($zip_o && $zip_o instanceof ZipArchive) {
			return $zip;
		}

		else {
			if ($zip->numFiles) {
				$zip->close();
				return $zip_file;
			}
			else {
				$zip->addEmptyDir(".");
				$zip->close();
				unlink($zip_file);
				return false;
			}
		}
	}
	public function files_and_folders($files = [], $folders = [], $options = []) {

		deprecated();

		if (are_empty($files, $folders))
			return false;

		$options = $this->core->array->merge(
			array(
				"name" => "files_and_folders",
				"path" => DROOT.FILES_FOLDER,

			),
			$options
		);

		$zip = new ZipArchive;

		$dir = sys_get_temp_dir();

		if (is_empty($dir))
			return false;

		$dir .= "/";
		$zip_file = $dir.key_of("name", $options).".zip";
		$result = $zip->open($zip_file, ZipArchive::CREATE | ZipArchive::OVERWRITE);

		if (!$result)
			return false;

		$this->files($files, $options, $zip);
		$this->folders($folders, $options, $zip);

		if ($zip->numFiles) {
			$zip->close();
			return $zip_file;
		}
		else {
			$zip->addEmptyDir(".");
			$zip->close();
			unlink($zip_file);
			return false;
		}

	}
	public function folders($folders = [],$options = [], $zip_o = false) {

		deprecated();

		if (is_empty($folders))
			return false;

		if (is_string($folders))
			$folders = array($folders);

		if (!is_array($folders))
			return false;

		$options = $this->core->array->merge(
			array(
				"name" => "folder",
				"path" => DROOT.FILES_FOLDER,

			),
			$options
		);

		if ($zip_o && $zip_o instanceof ZipArchive) {
			$zip = $zip_o;
		}

		else {

			$zip = new ZipArchive;

			$dir = sys_get_temp_dir();

			if (is_empty($dir))
				return false;

			$dir .= "/";
			$zip_file = $dir.key_of("name", $options).".zip";
			$result = $zip->open($zip_file, ZipArchive::CREATE | ZipArchive::OVERWRITE);

			if (!$result)
				return false;
		}

		foreach ($folders as $folder) {
				$zip->addEmptyDir($folder);
		}

		if ($zip_o && $zip_o instanceof ZipArchive) {
			return $zip;
		}

		else {
			if ($zip->numFiles) {
				$zip->close();
				return $zip_file;
			}
			else {
				$zip->addEmptyDir(".");
				$zip->close();
				unlink($zip_file);
				return false;
			}
		}
	}
	public function source_and_target($array = [], $options = [], $zip_o = false) {

		deprecated();

		if (is_empty($array))
			return false;

		$options = $this->core->array->merge(
			array(
				"name" => "files_and_folders",
				"folder" => false,
				"path" => DROOT.FILES_FOLDER,

			),
			$options
		);

		if ($zip_o && $zip_o instanceof ZipArchive) {
			$zip = $zip_o;
		}

		else {

			$zip = new ZipArchive;
			$dir = sys_get_temp_dir();

			if (is_empty($dir))
				return false;

			$dir .= "/";
			$zip_file = $dir.key_of("name", $options).".zip";
			$result = $zip->open($zip_file, ZipArchive::CREATE | ZipArchive::OVERWRITE);

			if (!$result)
				return false;
		}

		foreach ($array as $file) {

			$source = key_of("source", $file);
			$target = key_of("target", $file);

			if (any_empty($source, $target))
				continue;

			if (is_file($source))
				$zip->addFile($source, $target);
			else if (is_file(key_of("path", $options).$source))
				$zip->addFile(key_of("path", $options).$source, $target);
			else
				$zip->addFromString($target, $source);

			if (method_exists($zip, "setCompressionName"))
				$zip->setCompressionName($target, ZipArchive::CM_STORE);
		}

		if ($zip_o && $zip_o instanceof ZipArchive) {
			return $zip;
		}

		else {
			if ($zip->numFiles) {
				$zip->close();
				return $zip_file;
			}
			else {
				$zip->addEmptyDir(".");
				$zip->close();
				unlink($zip_file);
				return false;
			}
		}
	}

}

// Deprecated

class KB_ZipArchive extends KB_zip {

	function __construct () {

	}

}