<?php

$count = function($ID) use($core) {

	return $core->kit->data->count_per("group_ID", $ID);

};

$delete = function($ID) use ($core) {

	if (is_empty($ID))
		return "";

	if ($core->kit->data->count_per("group_ID", $ID) != 0)
		return "";

	$icon = $core->html->img("icon_waste.png");

	$return = $core->html->tag("a", [$icon, "Delete"], array(
		"href"=>$core->url(["do" => "delete", "ID" => $ID]),
		"class" => "red small button",
		"onclick" => "return confirm(\'Delete\')"
	));

	return $core->html->render($return);

};

$db = function($ID) use($core) {

	if (is_empty($ID))
		return "";

	$icon = $core->template->icon("waste");
	$label = $core->html->tag("span",
		$core->language->translate("Database")
	);

	$return = $core->html->tag("a", [$icon, $label], array(
		"href"=>$core->url(["do" => "delete_db", "ID" => $ID]),
		"class" => "red small button",
		"onclick" => "return confirm(\'Delete from DB\')"
	));

	return $core->html->render($return);

};

$description = function ($text = "", $file = "", $folder = "") use ($core) {

	if ($text == '$info_text') {

		if ($file == '$info_file')
			$text = "";

		else
			$text = $core->filesystem->read_file(KITS_FOLDER.$folder.'/'.$file);
	}

	$text = $core->language->translate($text);

	return $text;

};

$fs = function($folder) use($core) {

	if (is_empty($folder))
		return "";

	$icon = $core->template->icon("waste");
	$label = $core->html->tag("span",
		$core->language->translate("Filesystem")
	);

	$return = $core->html->tag("a", [$icon,$label], array(
		"href"=>$core->url(["do" => "delete_fs", "folder" => $folder]),
		"class" => "red small button",
		"onclick" => "return confirm(\'Delete from FS\')"
	));

	return $core->html->render($return);

};

$icon = function($icon) use($core) {

	if (is_empty($icon))
		return "";

	$file = base64_decode($icon);
	$f = finfo_open();
	$mime_type = finfo_buffer($f, $file, FILEINFO_MIME_TYPE);

	if (are_equal($mime_type, "text/plain"))
		$mime_type = "image/svg+xml";

	return $core->html->render(
		$core->html->img("data:".$mime_type.";base64, ".$icon, array("style" => "height: 1.2em"))
	);
};

$install = function($ID, $folder) use($core) {

	$icon = $core->template->icon("switch");

	if (is_empty($ID))
		$return = $core->html->tag("a",
			array(
				$icon,
				$core->html->tag("span", $core->language->translate("No"))
			),
			array(
				"href"=>$core->url(["do" => "install", "folder" => $folder]),
				"class" => "red small button"
			)
		);

	else
		$return = $core->html->tag("a",
			array(
				$icon,
				$core->html->tag("span", $core->language->translate("Yes"))
			),
			array(
				"href"=>$core->url(["do" => "deinstall", "ID" => $ID]),
				"class" => "green small button"
			)
		);

	return $core->html->render($return);

};

$local = function ($unique_name = "") use ($core) {

	if (is_empty($unique_name))
		return false;

	$kits = $core->kit->read_local_kits();

	$return = false;

	if ($match = fetch($unique_name,$kits)){

		$fs = $core->filesystem();
		$fs->dir($core->path("kits"));
		$file = $fs->file->read($match[key($match)]["folder"]."/version.txt");

		if($file)
			$return = int($file);
	}

	return $return;
};

$local_release = function($unique_name = "") use($core) {

	if (is_empty($unique_name))
		return false;

	$files = $core->file->find("release.php", DROOT.KITS_FOLDER);

	if (is_empty($files))
		return false;

	foreach ($files as $file) {

		$folder = $core->folder->name($file);

		$name = preg_replace(array(preg_mod(DROOT), preg_mod(KITS_FOLDER)),array("",""),$folder);
		$name = trim($name, "/");
		$name = preg_replace("-/-","_",$name);

		if (!are_equal($name, $unique_name))
			continue;

		$version = $core->file->read($folder."/version.txt");
		$version = int($version);
		return $version;


	}

	return false;

};

$name = function ($ID, $name) use ($core) {

	$name = $core->language->translate($name);

	return $name;

};

$edit = function ($ID) use ($core) {

	if (is_empty($ID))
		return;

	$return = $core->template->kit->button("edit","Edit",[
		"href" => $core->url(["do" => "edit", "ID" => $ID]),
		"class" => "small button",
	]);

	return $core->html->render($return);
};

$open = function ($ID) use($core) {
	$return = $core->html->tag("a",$core->language->translate("open"), ["class" => "small button", "href"=>$core->url(["do" => "kitinfo", "ID" => $ID])]);
	return $core->html->render($return);
};

$releases = function($kits) use($core) {

	if (any_empty($kits, is_indexed($kits)))
		return array();

	foreach ($kits as $index => $kit) {

		$path = "";
		$paths = explode("/",key_of("folder", $kit));

		foreach ($paths as $folder) {
			$path = $path.$folder."/";

			if ($core->file->exists(DROOT.KITS_FOLDER.$path."release.php")) {

				unset($kits[$index]);

				$config = $core->file->read_php(DROOT.KITS_FOLDER.$path."release.php", "", "config");

				if (is_empty($config))
					break;

				if (not_empty(fetch(array("folder"=>$path), $kits)))
					break;

				$kits[] = array(
					"name" => key_of("title", $config),
					"folder" => $path,
				);

				break;

			}
		}

	}

	return $kits;

};

$update = function ($ID) use($core) {

	if (is_empty($ID))
		return "";

	$icon = $core->template->icon("switch");
	$label = $core->html->tag("span",
		$core->language->translate("Update Configuration")
	);

	$return = $core->html->tag("a",[$icon,$label], [
		"href" => $core->url(["do" => "reload", "ID" => $ID]),
		"class" => "small button",
	]);

	return $core->html->render($return);
};

$local_toolbar = function ($ID = "", $folder = "") use ($core, $update, $db, $fs, $edit) {

	return $core->html->render([
		$edit($ID),
		$update($ID),
		$db($ID),
		$fs($folder),
	]);

};

$groups_toolbar = function ($ID = "") use ($core, $delete, $edit) {

	return $core->html->render([
		$edit($ID),
		$delete($ID),
	]);

};
