<?php

$kits_fs = $core->filesystem($core->path("kits"));

$table = $core->view->table();

$table->new_columns(array(
	"name" => array(
		"title" => "Name"
	),
	"version" => array(
		"title" => "Version"
	),
	"version_remote" => array(
		"title" => "Version available"
	),
	"update" => array(
		"title" => "Update",
		"value" => 'return_update($remote_kit_ID)',
		"markup" => 'style="width:0"',
	),
));

$table->new_row("remarks",[
	"class"=> "hide"
]);

$return_update = function($ID) use ($core) {

	if (any_empty($ID, is_numeric($ID)))
		return "";

	return $core->template->kit->small_button("download", "", ["href"=>$core->url(["do"=>"download", "ID"=>$ID])]);
};

$table->config("return_update", $return_update);


$url = "http://kitbooth.com/pages/kit-store-public-kits.html?framework=".$core->version("framework")."&core=".$core->version("compatibility");
$store_kits = $core->file->read($url);
$store_kits = $core->json->decode($store_kits);

$db_kits = $core->kit->data->read();

foreach ($db_kits as $key => $db_kit) {

	$version = $kits_fs->file->read(key_of("folder", $db_kit)."/version.txt");

	if (is_empty($version)) {
		unset($db_kits[$key]);
		continue;
	}


	$remote_kit = either(
		$core->array->filter($store_kits, ["name"=>key_of("unique_name", $db_kit)]),
		$core->array->filter($private_kits, ["name"=>key_of("unique_name", $db_kit)])
	);

	$version_remote = key_fetch("version", $remote_kit);

	if (
		is_empty($version_remote)
		|| are_equal($version, $version_remote)
	) {
		unset($db_kits[$key]);
		continue;
	}

	$remote_kit_ID = key_fetch("ID", $remote_kit);
	$remarks = key_fetch("remarks_all", $remote_kit);

	$db_kits[$key]["remarks"] = "";

	foreach ($remarks as $remark_version => $remark) {
		if ($remark_version <= $version)
			continue;

		$db_kits[$key]["remarks"] .= $core->html->render([
			$core->html->h2($remark_version),
			$core->markdown->decode($remark, ["shift"=>2])
		]);

	}

	$db_kits[$key]["version"] = $version;
	$db_kits[$key]["version_remote"] = $version_remote;
	$db_kits[$key]["remote_kit_ID"] = $remote_kit_ID;

}

echo $table->render($db_kits);
